# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from pyworkflow.tests import *

from pwem import Domain
import pwem.protocols as emprot

from .test_workflow import TestWorkflow


class TestMixedRelionTutorial(TestWorkflow):
    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)
        cls.dataset = DataSet.getDataSet('relion_tutorial')
        cls.crdsEmanDir = cls.dataset.getFile('boxingDir')
        cls.crdsXmippDir = cls.dataset.getFile('posAllDir')
        cls.micsFn = cls.dataset.getFile('allMics')
        cls.vol = cls.dataset.getFile('volume')

    def test_workflow(self):
        # First, import a set of micrographs
        print("Importing a set of micrographs...")
        protImport = self.newProtocol(emprot.ProtImportMicrographs,
                                      filesPath=self.micsFn,
                                      samplingRateMode=1, magnification=79096,
                                      scannedPixelSize=56, voltage=300,
                                      sphericalAberration=2.0)
        protImport.setObjLabel('import 20 mics')
        self.launchProtocol(protImport)
        self.assertIsNotNone(protImport.outputMicrographs,
                             "There was a problem with the import")

        print("Importing a volume...")
        protImportVol = self.newProtocol(emprot.ProtImportVolumes, filesPath=self.vol,
                                         samplingRate=7.08)
        protImportVol.setObjLabel('import single vol')
        self.launchProtocol(protImportVol)
        self.assertIsNotNone(protImportVol.outputVolume, "There was a problem with the import")

        print("Preprocessing the micrographs...")
        xmippProtocols = Domain.importFromPlugin('xmipp3.protocols',
                                                 doRaise=True)
        protPreprocess = self.newProtocol(xmippProtocols.XmippProtPreprocessMicrographs,
                                          doCrop=True, cropPixels=25)
        protPreprocess.inputMicrographs.set(protImport.outputMicrographs)
        protPreprocess.setObjLabel('crop 50px')
        self.launchProtocol(protPreprocess)
        self.assertIsNotNone(protPreprocess.outputMicrographs,
                             "There was a problem with the downsampling")

        print("Running Eman import coordinates...")
        protPP = self.newProtocol(emprot.ProtImportCoordinates,
                                  importFrom=emprot.ProtImportCoordinates.IMPORT_FROM_EMAN,
                                  filesPath=self.crdsEmanDir,
                                  filesPattern='info/*_info.json', boxSize=60)
        protPP.inputMicrographs.set(protPreprocess.outputMicrographs)
        protPP.setObjLabel('import from Eman boxing')
        self.launchProtocol(protPP)
        self.assertIsNotNone(protPP.outputCoordinates,
                             "There was a problem with the Eman import coordinates")

        # Now estimate CTF on the micrographs with ctffind 
        print("Performing CTFfind...")
        grigorieffLabProtocols = Domain.importFromPlugin(
            'cistem.protocols',
            doRaise=True)
        protCTF = self.newProtocol(grigorieffLabProtocols.CistemProtCTFFind,
                                   minDefocus=12000, maxDefocus=30000,
                                   numberOfThreads=16)
        protCTF.inputMicrographs.set(protPreprocess.outputMicrographs)
        protCTF.setObjLabel('CTF ctffind')
        self.launchProtocol(protCTF)

        print("Run extract particles with <Same as picking> option")
        protExtract = self.newProtocol(xmippProtocols.XmippProtExtractParticles,
                                       boxSize=60,
                                       downsampleType=emprot.SAME_AS_PICKING,
                                       doRemoveDust=False,
                                       doInvert=False,
                                       doFlip=False, backRadius=28, runMode=1)
        protExtract.inputCoordinates.set(protPP.outputCoordinates)
        protExtract.ctfRelations.set(protCTF.outputCTF)
        protExtract.setObjLabel('Extract particles')
        self.launchProtocol(protExtract)
        self.assertIsNotNone(protExtract.outputParticles,
                             "There was a problem with the extract particles")

        print("Run CL2D")
        protCL2D = xmippProtocols.XmippProtCL2D(numberOfClasses=32,
                                                numberOfInitialClasses=4,
                                                numberOfIterations=2,
                                                numberOfMpi=16)
        protCL2D.inputParticles.set(protExtract.outputParticles)
        protCL2D.setObjLabel('CL2D')
        self.launchProtocol(protCL2D)
        self.assertIsNotNone(protCL2D.outputClasses,
                             "There was a problem with CL2D")

        # Now estimate CTF on the micrographs with xmipp
        print("Performing Xmipp CTF...")
        protCTF2 = self.newProtocol(xmippProtocols.XmippProtCTFMicrographs,
                                    doInitialCTF=True,
                                    minDefocus=12000, maxDefocus=30000,
                                    runMode=1, numberOfMpi=1, numberOfThreads=16)
        protCTF2.ctfRelations.set(protCTF.outputCTF)
        protCTF2.inputMicrographs.set(protPreprocess.outputMicrographs)
        protCTF2.setObjLabel('CTF xmipp')
        self.launchProtocol(protCTF2)

        print("Running Xmipp Import Coordinates")
        protPP2 = self.newProtocol(emprot.ProtImportCoordinates,
                                   importFrom=emprot.ProtImportCoordinates.IMPORT_FROM_XMIPP,
                                   filesPath=self.crdsXmippDir,
                                   filesPattern='*.pos', boxSize=60)
        protPP2.inputMicrographs.set(protPreprocess.outputMicrographs)
        protPP2.setObjLabel('Xmipp Import Coords')
        self.launchProtocol(protPP2)
        self.assertIsNotNone(protPP2.outputCoordinates,
                             "There was a problem with the Xmipp import coordinates")

        print("Run extract particles with <Same as picking> option")
        protExtract2 = self.newProtocol(xmippProtocols.XmippProtExtractParticles,
                                        boxSize=60,
                                        downsampleType=emprot.SAME_AS_PICKING,
                                        doRemoveDust=False, doInvert=True,
                                        doFlip=False, backRadius=28, runMode=1)
        protExtract2.inputCoordinates.set(protPP2.outputCoordinates)
        protExtract2.ctfRelations.set(protCTF2.outputCTF)
        protExtract2.setObjLabel('Extract particles')
        self.launchProtocol(protExtract2)
        self.assertIsNotNone(protExtract2.outputParticles,
                             "There was a problem with the extract particles")

        print("Run Relion Classification2d")
        relionProtocols = Domain.importFromPlugin('relion.protocols',
                                                  doRaise=True)
        prot2D = relionProtocols.ProtRelionClassify2D(numberOfMpi=1,
                                                      numberOfThreads=4)
        prot2D.numberOfClasses.set(50)
        prot2D.inputParticles.set(protExtract2.outputParticles)
        prot2D.setObjLabel('relion 2D')
        self.launchProtocol(prot2D)
        self.assertIsNotNone(prot2D.outputClasses,
                             "There was a problem with Relion 2D:\n" + (prot2D.getErrorMessage() or "No error set"))

        print("Run Relion Refine")
        proRef = relionProtocols.ProtRelionRefine3D()
        proRef.inputParticles.set(protExtract2.outputParticles)
        proRef.referenceVolume.set(protImportVol.outputVolume)
        proRef.setObjLabel('relion Refine')
        self.launchProtocol(proRef)
        self.assertIsNotNone(proRef.outputVolume,
                             "There was a problem with Relion Refine:\n" + (proRef.getErrorMessage() or "No error set"))
        self.assertIsNotNone(proRef.outputParticles,
                             "There was a problem with Relion Refine:\n" + (proRef.getErrorMessage() or "No error set"))


class TestMixedFrealignClassify(TestWorkflow):
    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)
        cls.dataset = DataSet.getDataSet('relion_tutorial')
        cls.crdsEmanDir = cls.dataset.getFile('boxingDir')
        cls.crdsXmippDir = cls.dataset.getFile('posAllDir')
        cls.micsFn = cls.dataset.getFile('allMics')
        cls.vol = cls.dataset.getFile('volume')

    def test_workflow(self):
        # First, import a set of micrographs
        print("Importing a set of micrographs...")
        # TODO missing contrst amplitude ROB
        protImport = self.newProtocol(emprot.ProtImportMicrographs,
                                      filesPath=self.micsFn,
                                      samplingRateMode=1, magnification=79096,
                                      scannedPixelSize=56, voltage=300,
                                      sphericalAberration=2.0)
        protImport.setObjLabel('import 20 mics')
        self.launchProtocol(protImport)
        self.assertIsNotNone(protImport.outputMicrographs,
                             "There was a problem with the import")

        print("Importing a volume...")
        protImportVol = self.newProtocol(emprot.ProtImportVolumes,
                                         filesPath=self.vol, samplingRate=7.08)
        protImportVol.setObjLabel('import single vol')
        self.launchProtocol(protImportVol)
        self.assertIsNotNone(protImportVol.outputVolume,
                             "There was a problem with the import")

        print("Preprocessing the micrographs...")
        xmippProtocols = Domain.importFromPlugin('xmipp3.protocols',
                                                 doRaise=True)
        protPreprocess = self.newProtocol(xmippProtocols.XmippProtPreprocessMicrographs,
                                          doCrop=True, cropPixels=25)
        protPreprocess.inputMicrographs.set(protImport.outputMicrographs)
        protPreprocess.setObjLabel('crop 50px')
        self.launchProtocol(protPreprocess)
        self.assertIsNotNone(protPreprocess.outputMicrographs,
                             "There was a problem with the downsampling")

        print("Running Eman import coordinates...")
        protPP = self.newProtocol(emprot.ProtImportCoordinates,
                                  importFrom=emprot.ProtImportCoordinates.IMPORT_FROM_EMAN,
                                  filesPath=self.crdsEmanDir,
                                  filesPattern='info/*_info.json', boxSize=60)
        protPP.inputMicrographs.set(protPreprocess.outputMicrographs)
        protPP.setObjLabel('import from Eman boxing')
        self.launchProtocol(protPP)
        self.assertIsNotNone(protPP.outputCoordinates,
                             "There was a problem with the Eman import coordinates")

        # Now estimate CTF on the micrographs with ctffind 
        print("Performing CTFfind...")
        grigorieffLabProtocols = Domain.importFromPlugin(
            'cistem.protocols',
            doRaise=True)
        protCTF = self.newProtocol(grigorieffLabProtocols.CistemProtCTFFind,
                                   minDefocus=12000, maxDefocus=30000,
                                   numberOfThreads=16)
        protCTF.inputMicrographs.set(protPreprocess.outputMicrographs)
        protCTF.setObjLabel('CTF ctffind')
        self.launchProtocol(protCTF)

        print("Run extract particles with <Same as picking> option")
        protExtract = self.newProtocol(xmippProtocols.XmippProtExtractParticles,
                                       boxSize=60,
                                       downsampleType=emprot.SAME_AS_PICKING,
                                       doRemoveDust=False,
                                       doInvert=False,
                                       doFlip=False, backRadius=28, runMode=1)
        protExtract.inputCoordinates.set(protPP.outputCoordinates)
        protExtract.ctfRelations.set(protCTF.outputCTF)
        protExtract.setObjLabel('Extract particles')
        self.launchProtocol(protExtract)
        self.assertIsNotNone(protExtract.outputParticles,
                             "There was a problem with the extract particles")



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *              Vahid Abrishami (vabrishami@cnb.csic.es)
# *              Josue Gomez Blanco (josue.gomez-blanco@mcgill.ca)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import enum
import os
import warnings

try:
    from itertools import izip
except ImportError:
    izip = zip
from math import ceil

import pyworkflow.object as pwobj
import pyworkflow.utils as pwutils
from pyworkflow.gui.plotter import Plotter
import pyworkflow.protocol.params as params
import pyworkflow.protocol.constants as pwcts

import pwem.objects as emobj
from pwem import emlib

from pwem.protocols import ProtProcessMovies

OUT_MICS = "outputMicrographs"
OUT_MICS_DW = "outputMicrographsDoseWeighted"
OUT_MOVIES = "outputMovies"
OUT_MICS_ODD = 'outputMicrographsOdd'
OUT_MICS_EVEN = 'outputMicrographsEven'


class ProtAlignMovies(ProtProcessMovies):
    """
    Base class for movie alignment protocols such as:
    motioncorr, crosscorrelation and optical flow

    Alignment parameters are defined in common. For example,
    the frames range used for alignment and final sum, the binning factor
    or the cropping options (region of interest)
    """
    _possibleOutputs = {OUT_MICS: emobj.SetOfMicrographs,
                        OUT_MICS_DW: emobj.SetOfMicrographs,
                        OUT_MICS_ODD: emobj.SetOfMicrographs,
                        OUT_MICS_EVEN: emobj.SetOfMicrographs,
                        OUT_MOVIES: emobj.SetOfMovies}

    # Even / odd functionality
    evenOddCapable = False

    # -------------------------- DEFINE param functions -----------------------
    def _defineParams(self, form):
        ProtProcessMovies._defineParams(self, form)
        self._defineAlignmentParams(form)

    def _defineAlignmentParams(self, form):
        group = form.addGroup('Alignment')
        line = group.addLine('Frames to ALIGN',
                             help='Frames range to ALIGN on each movie. The '
                                  'first frame is 1. If you set 0 in the final '
                                  'frame to align, it means that you will '
                                  'align until the last frame of the movie.')
        line.addParam('alignFrame0', params.IntParam, default=1,
                      label='from')
        line.addParam('alignFrameN', params.IntParam, default=0,
                      label='to')
        group.addParam('useAlignToSum', params.BooleanParam, default=True,
                       label='Use ALIGN frames range to SUM?',
                       help="If *Yes*, the same frame range will be used to "
                            "ALIGN and to SUM. If *No*, you can selected a "
                            "different range for SUM (must be a subset).")
        line = group.addLine('Frames to SUM', condition="not useAlignToSum",
                             help='Frames range to SUM on each movie. The '
                                  'first frame is 1. If you set 0 in the final '
                                  'frame to sum, it means that you will sum '
                                  'until the last frame of the movie.')
        line.addParam('sumFrame0', params.IntParam, default=1,
                      label='from')
        line.addParam('sumFrameN', params.IntParam, default=0,
                      label='to')
        group.addParam('binFactor', params.FloatParam, default=1.,
                       label='Binning factor',
                       help='1x or 2x. Bin stack before processing.')

        line = group.addLine('Crop offsets (px)')
        line.addParam('cropOffsetX', params.IntParam, default=0, label='X')
        line.addParam('cropOffsetY', params.IntParam, default=0, label='Y')

        line = group.addLine('Crop dimensions (px)',
                             help='How many pixels to crop from offset\n'
                                  'If equal to 0, use maximum size.')
        line.addParam('cropDimX', params.IntParam, default=0, label='X')
        line.addParam('cropDimY', params.IntParam, default=0, label='Y')

        form.addParam('doSaveAveMic', params.BooleanParam, default=True,
                      label="Save aligned micrograph",
                      expertLevel=pwcts.LEVEL_ADVANCED)

        form.addParam('doSaveMovie', params.BooleanParam, default=False,
                      label="Save movie", expertLevel=pwcts.LEVEL_ADVANCED,
                      help="Save Aligned movie")

        if self.evenOddCapable:
            form.addParam('splitEvenOdd', params.BooleanParam,
                          default=False,
                          label='Split & sum odd/even frames?',
                          expertLevel=params.LEVEL_ADVANCED,
                          help='Generate odd and even sums using odd and even frames '
                               'respectively when this option is enabled.')

    # --------------------------- STEPS functions ----------------------------
    def createOutputStep(self):
        # validate that we have some output movies
        for outName, out in self.iterOutputAttributes():
            output = out
            break

        if output.getSize() == 0 and len(self.listOfMovies) != 0:
            raise Exception("All movies failed, didn't create outputMicrographs."
                            "Please review movie processing steps above.")
        elif output.getSize() < len(self.listOfMovies):
            self.warning(pwutils.yellowStr("WARNING - Failed to align %d movies."
                                           % (len(self.listOfMovies) - output.getSize())))

    def _loadOutputSet(self, SetClass, baseName, fixSampling=True):
        """
        Load the output set if it exists or create a new one.
        fixSampling: correct the output sampling rate if binning was used,
        except for the case when the original movies are kept and shifts
        refers to that one.
        """
        setFile = self._getPath(baseName)

        if os.path.exists(setFile) and os.path.getsize(setFile) > 0:
            outputSet = SetClass(filename=setFile)
            outputSet.loadAllProperties()
            outputSet.enableAppend()
        else:
            outputSet = SetClass(filename=setFile)
            outputSet.setStreamState(outputSet.STREAM_OPEN)

            inputMovies = self.inputMovies.get()
            outputSet.copyInfo(inputMovies)

            if fixSampling:
                newSampling = inputMovies.getSamplingRate() * self._getBinFactor()
                outputSet.setSamplingRate(newSampling)

        return outputSet

    def _updateOutputMicSet(self, newDone, sqliteFn, getOutputMicName,
                            outputName, streamMode):
        """ Updated the output micrographs set with new items found. """
        micSet = self._loadOutputSet(emobj.SetOfMicrographs, sqliteFn)
        doneFailed = []

        for movie in newDone:
            mic = micSet.ITEM_TYPE()
            mic.copyObjId(movie)
            mic.setMicName(movie.getMicName())
            # The subclass protocol is responsible for generating the output
            # micrograph file in the extra path with the required name
            extraMicFn = self._getExtraPath(getOutputMicName(movie))
            mic.setFileName(extraMicFn)
            if not os.path.exists(extraMicFn):
                print(pwutils.yellowStr("WARNING: Micrograph %s was not generated, "
                                        "can't add it to output set." % extraMicFn))
                doneFailed.append(movie)
                continue
            # Tolerate errors here. Usually here some plots are generated.
            try:
                self._preprocessOutputMicrograph(mic, movie)
            except Exception as e:
                self.error("Couldn't prepare output details: %s" % e)
                doneFailed.append(movie)
                continue

            micSet.append(mic)

        self._updateOutputSet(outputName, micSet, streamMode)
        if doneFailed:
            self._writeFailedList(doneFailed)

        if self._firstTimeOutput:
            # We consider that Movies are 'transformed' into the Micrographs
            # This will allow to extend the CTF associated to a set of
            # micrographs to another set of micrographs generated from a
            # different movie alignment
            self._defineTransformRelation(self.inputMovies, micSet)

    def _updateOutputMovieSet(self, newDone, streamMode):
        saveMovie = self.getAttributeValue('doSaveMovie', False)
        movieSet = self._loadOutputSet(emobj.SetOfMovies, 'movies.sqlite',
                                       fixSampling=saveMovie)

        # If we need to save the movies
        if saveMovie:
            movieSet.setGain(None)
            movieSet.setDark(None)

        for movie in newDone:
            try:
                newMovie = self._createOutputMovie(movie)
                if newMovie.getAlignment().getShifts()[0]:
                    movieSet.append(newMovie)
                else:
                    print(pwutils.yellowStr("WARNING: Movie %s has empty alignment "
                                            "data, can't add it to output set."
                                            % movie.getFileName()))

            # Warn about any exception creating the movie
            except Exception as e:
                print(pwutils.redStr("ERROR: Movie %s couldn't be "
                                     "added to the output set.\n%s"
                                     % (movie.getFileName(), e)))

        self._updateOutputSet(OUT_MOVIES, movieSet, streamMode)

        if self._firstTimeOutput:
            # Probably is a good idea to store a cached summary for the
            # first resulting movie of the processing.
            self._storeSummary(newDone[0])
            # If the movies are not written out, then dimensions can be
            # copied from the input movies
            if not saveMovie:
                movieSet.setDim(self.inputMovies.get().getDim())
            self._defineTransformRelation(self.inputMovies, movieSet)

    def _updateOutputSets(self, newDone, streamMode):
        if self._createOutputMovies():
            self._updateOutputMovieSet(newDone, streamMode)

        if self._createOutputMicrographs():
            self._updateOutputMicSet(newDone, 'micrographs.sqlite',
                                self._getOutputMicName,
                                OUT_MICS, streamMode)

        if self._createOutputWeightedMicrographs():
            self._updateOutputMicSet(newDone, 'micrographs_dose-weighted.sqlite',
                                self._getOutputMicWtName,
                                OUT_MICS_DW, streamMode)

        if self._doSplitEvenOdd():
            self._updateOutputMicSet(newDone, 'micrographs_even.sqlite',
                                self._getOutputMicEvenName,
                                OUT_MICS_EVEN, streamMode)
            self._updateOutputMicSet(newDone, 'micrographs_odd.sqlite',
                                self._getOutputMicOddName,
                                OUT_MICS_ODD, streamMode)

    def _checkNewOutput(self):
        if getattr(self, 'finished', False):
            return

        # Load previously done items (from text file)
        doneList = self._readDoneList()
        # Check for newly done items
        newDone = [m for m in self.listOfMovies
                   if m.getObjId() not in doneList and self._isMovieDone(m)]

        # Update the file with the newly done movies
        # or exit from the function if no new done movies
        self.debug('_checkNewOutput: ')
        self.debug('   listOfMovies: %s, doneList: %s, newDone: %s'
                   % (len(self.listOfMovies), len(doneList), len(newDone)))

        self._firstTimeOutput = len(doneList) == 0
        allDone = len(doneList) + len(newDone)
        # We have finished when there is not more input movies (stream closed)
        # and the number of processed movies is equal to the number of inputs
        self.finished = self.streamClosed and allDone == len(self.listOfMovies)
        streamMode = pwobj.Set.STREAM_CLOSED if self.finished else pwobj.Set.STREAM_OPEN

        if newDone:
            self._writeDoneList(newDone)

        elif not self.finished:
            # If we are not finished and no new output have been produced
            # it does not make sense to proceed and updated the outputs
            # so we exit from the function here
            return

        self.debug('   finished: %s ' % self.finished)
        self.debug('        self.streamClosed (%s) AND' % self.streamClosed)
        self.debug('        allDone (%s) == len(self.listOfMovies (%s)'
                   % (allDone, len(self.listOfMovies)))
        self.debug('   streamMode: %s' % streamMode)

        self._updateOutputSets(newDone, streamMode)

        if self.finished:  # Unlock createOutputStep if finished all jobs
            outputStep = self._getFirstJoinStep()
            if outputStep and outputStep.isWaiting():
                outputStep.setStatus(pwcts.STATUS_NEW)

    # --------------------------- INFO functions ------------------------------
    def _validate(self):
        errors = []

        # Only validate about cropDimensions if the protocol supports them
        if (hasattr(self, 'cropDimX') and hasattr(self, 'cropDimY')
                and (self.cropDimX > 0 >= self.cropDimY
                     or self.cropDimX <= 0 < self.cropDimY)):
            errors.append("If you give cropDimX, you should also give "
                          "cropDimY and vice versa")

        inputMovies = self.inputMovies.get()

        # Do not continue if there ar no movies. Validation message will
        # take place since attribute is a Pointer.
        if inputMovies is None:
            return errors

        firstItem = inputMovies.getFirstItem()

        firstFrame, lastFrame, _ = inputMovies.getFramesRange()
        if lastFrame == 0:
            # Although getFirstItem is not recommended in general, here it is
            # used only once, for validation purposes, so performance
            # problems should not appear.
            frames = firstItem.getNumberOfFrames()
            lastFrame = frames
        else:
            frames = lastFrame - firstFrame + 1

        if frames is not None:
            def _validateRange(prefix):
                # Avoid validation when the range is not defined
                if not hasattr(self, '%sFrame0' % prefix):
                    return

                f0, fN = self._getFrameRange(frames, prefix)
                if fN < firstFrame or fN > lastFrame:
                    errors.append("Check the selected last frame to *%s*. "
                                  "Last frame (%d) should be in range: %s "
                                  % (prefix.upper(), fN, (firstFrame,
                                                          lastFrame)))
                if f0 < firstFrame or f0 > lastFrame:
                    errors.append("Check the selected first frame to *%s*. "
                                  "First frame (%d) should be in range: %s "
                                  % (prefix.upper(), f0, (firstFrame,
                                                          lastFrame)))
                if fN < f0:
                    errors.append("Check the selected frames range to *%s*. "
                                  "Last frame (%d) should be greater or equal "
                                  "than first frame (%d)"
                                  % (prefix.upper(), fN, f0))

            _validateRange("align")
            _validateRange("sum")

        if not emlib.image.ImageHandler().existsLocation(firstItem.getLocation()):
            errors.append("The input movie files do not exist!!! "
                          "Since usually input movie files are symbolic links, "
                          "please check that links are not broken if you "
                          "moved the project folder. ")

        return errors

    # --------------------------- INFO functions -------------------------------
    def _summary(self):
        return [self.summaryVar.get('')]

    # --------------------------- UTILS functions ----------------------------
    def _doSplitEvenOdd(self):
        """ Returns if even/odd stuff has to be done"""
        if not self.evenOddCapable:
            return False
        else:
            return self.splitEvenOdd.get()

    def _useAlignToSum(self):
        return self.getAttributeValue('useAlignToSum', False)

    def _getFrameRange(self, n, prefix):
        """
        Params:
        :param n: Number of frames of the movies
        :param prefix: what range we want to consider, either 'align' or 'sum'
        :return: (i, f) initial and last frame range
        """
        # In case that the user select the same range for ALIGN and SUM
        # we also use the 'align' prefix
        if self._useAlignToSum():
            prefix = 'align'

        first = self.getAttributeValue('%sFrame0' % prefix)
        last = self.getAttributeValue('%sFrameN' % prefix)

        if first <= 1:
            first = 1

        if last <= 0:
            last = n

        return first, last

    def _createOutputMovie(self, movie):
        # Parse the alignment parameters and store the log files
        alignedMovie = movie.clone()
        n = movie.getNumberOfFrames()
        first, last = self._getFrameRange(n, 'align')
        framesRange = alignedMovie.getFramesRange()
        framesRange.setFirstFrame(first)
        framesRange.setLastFrame(last)
        # Check if user selected to save movie, use the getAttributeValue
        # function for allow the protocol to not define this flag
        # and use False as default
        if self.getAttributeValue('doSaveMovie', False):
            # The subclass protocol is responsible for generating the output
            # movie file in the extra path with the required name
            extraMovieFn = self._getExtraPath(self._getOutputMovieName(movie))
            alignedMovie.setFileName(extraMovieFn)
            # When the output movies are saved, the shifts
            # will be set to zero since they are aligned
            totalFrames = last - first + 1
            xshifts = [0] * totalFrames
            yshifts = xshifts
            # If we save the movies, we need to modify which are the index
            # of the first frame in the stack, now is 1 since the stack is
            # written only with the given frames
            firstFrameIndex = 1
        else:
            xshifts, yshifts = self._getMovieShifts(movie)
            firstFrameIndex = first

        framesRange.setFirstFrameIndex(firstFrameIndex)
        alignment = emobj.MovieAlignment(first=first, last=last, xshifts=xshifts,
                                         yshifts=yshifts)

        roiList = [self.getAttributeValue(s, 0) for s in
                   ['cropOffsetX', 'cropOffsetY', 'cropDimX', 'cropDimY']]
        alignment.setRoi(roiList)
        alignedMovie.setAlignment(alignment)

        return alignedMovie

    # ---------- Hook functions that need to be implemented in subclasses ------
    def _getBinFactor(self):
        return self.getAttributeValue('binFactor', 1.0)

    def _getMovieRoot(self, movie):
        # Try to use the 'original' fileName in case it is present
        # the original could be different from the current filename if
        # we are dealing with compressed movies (e.g., movie.mrc.bz2)
        fn = movie.getAttributeValue('_originalFileName',
                                     movie.getFileName())
        # Remove the first extension
        fnRoot = pwutils.removeBaseExt(fn)
        # Check if there is a second extension
        # (Assuming it is only a dot and 3 or 4 characters after it
        # Do not perform this check if the file name is short
        if len(fnRoot) > 5:
            if fnRoot[-4] == '.' or fnRoot[-5] == '.':
                fnRoot = pwutils.removeExt(fnRoot)

        return fnRoot

    def _getOutputMovieName(self, movie):
        """ Returns the name of the output movie.
        (relative to micFolder)
        """
        return self._getMovieRoot(movie) + '_aligned_movie.mrcs'

    def _getOutputMicName(self, movie):
        """ Returns the name of the output micrograph
        (relative to micFolder)
        """
        return self._getMovieRoot(movie) + '_aligned_mic.mrc'

    def _getOutputMicWtName(self, movie):
        """ Returns the name of the output dose-weighted micrograph
        (relative to micFolder)
        """
        return self._getMovieRoot(movie) + '_aligned_mic_DW.mrc'

    def _getOutputMicEvenName(self, movie):
        """ Returns the name of the output EVEN micrograph
        (relative to micFolder)
        """
        return self._getMovieRoot(movie) + '_aligned_mic_EVN.mrc'

    def _getOutputMicOddName(self, movie):
        """ Returns the name of the output EVEN micrograph
        (relative to micFolder)
        """
        return self._getMovieRoot(movie) + '_aligned_mic_ODD.mrc'

    def _getOutputMicThumbnail(self, movie):
        return self._getExtraPath(self._getMovieRoot(movie) + '_thumbnail.png')

    def _getMovieShifts(self, movie):
        """ Returns the x and y shifts for the alignment of this movie.
         The shifts should refer to the original micrograph without any binning.
         In case of a binning greater than 1, the shifts should be scaled.
        """
        return [], []

    def _createOutputMovies(self):
        """ Returns True if an output set of movies will be generated.
        The most common case is to always generate output movies,
        either with alignment only or the binary aligned movie files.
        Subclasses can override this function to change this behavior.
        """
        return True

    def _createOutputMicrographs(self):
        """ By default check if the user have selected 'doSaveAveMic'
        property. Subclasses can override this method to implement different
        behaviour.
        """
        return self.getAttributeValue('doSaveAveMic', True)

    def _createOutputWeightedMicrographs(self):
        return False

    def _preprocessOutputMicrograph(self, mic, movie):
        """ Hook function that will be call before adding the micrograph
        to the output set of micrographs.
        """
        pass

    def _doComputeMicThumbnail(self):
        """ Should be implemented in sub-classes if want to check
        the generation of thumbnails.
        """
        return False

    def _storeSummary(self, movie):
        """ Implement this method if you want to store the summary. """
        pass

    def _getCorrectedDose(self, movieSet):
        """get and correct the pre-exposure dose. It is important for cases
        in which the first frame is different of one. The method support both
        movie and sets of movies"""

        firstFrame, _, _ = movieSet.getFramesRange()
        preExp = movieSet.getAcquisition().getDoseInitial()
        dose = movieSet.getAcquisition().getDosePerFrame()
        preExp += dose * (firstFrame - 1)

        return preExp, dose

    def __runXmippProgram(self, program, args):
        """ Internal shortcut function to launch a Xmipp program. """
        from pwem import Domain
        xmipp3 = Domain.importFromPlugin('xmipp3')
        xmipp3.Plugin.runXmippProgram(program, args)

    def __runEman2Program(self, program, args):
        """ Internal workaround to launch an EMAN2 program. """
        from pwem import Domain
        eman2 = Domain.importFromPlugin('eman2')
        from pyworkflow.utils.process import runJob
        runJob(self._log, eman2.Plugin.getProgram(program), args,
               env=eman2.Plugin.getEnviron())

    def computePSD(self, inputMic, oroot, dim=384,  # 384 = 128 + 256, which should be fast for any Fourier Transformer
                   overlap=0.4):
        warnings.warn("Use psd = image.computePSD(overlap=0.4, xdim=384, ydim=384, fftthreads=1) instead",
                      DeprecationWarning)
        ih = emlib.image.ImageHandler()
        psdImg1 = ih.read(inputMic)
        res = psdImg1.computePSD(overlap, dim, dim)
        res.write(oroot + ".psd")

    def composePSDImages(self, psdImg1, psdImg2, outputFn,
                         outputFnUncorrected=None, outputFnCorrected=None):
        """ Compose a single PSD image:
         left part from psd1 (uncorrected PSD),
         right-part from psd2 (corrected PSD)
        """
        data1 = psdImg1.getData()  # get the data now, as conversion would change them
        if outputFnUncorrected is not None:
            psdImg1.convertPSD()
            psdImg1.write(outputFnUncorrected)

        data2 = psdImg2.getData()  # get the data now, as conversion would change them
        if outputFnCorrected is not None:
            psdImg2.convertPSD()
            psdImg2.write(outputFnCorrected)

        # Compute middle index
        x, _, _, _ = psdImg1.getDimensions()
        m = int(round(x / 2.))
        data1[:, :m] = data2[:, :m]
        psdImg1.setData(data1)
        psdImg1.write(outputFn)

    def composePSD(self, psd1, psd2, outputFn,
                   outputFnUncorrected=None, outputFnCorrected=None):
        import warnings
        warnings.warn("Use composePSDImages() instead", DeprecationWarning)
        """ Compose a single PSD image:
         left part from psd1 (uncorrected PSD),
         right-part from psd2 (corrected PSD)
        """
        ih = emlib.image.ImageHandler()
        self.composePSDImages(ih.read(psd1), ih.read(psd2), outputFn,
                              outputFnUncorrected, outputFnCorrected)

    def computePSDImages(self, movie, fnUncorrected, fnCorrected,
                         outputFnUncorrected=None, outputFnCorrected=None):
        self.composePSDImages(
            emlib.image.ImageHandler().read(fnUncorrected).computePSD(),
            emlib.image.ImageHandler().read(fnCorrected).computePSD(),
            self._getPsdCorr(movie),
            outputFnUncorrected,
            outputFnCorrected)

    def computePSDs(self, movie, fnUncorrected, fnCorrected,
                    outputFnUncorrected=None, outputFnCorrected=None):
        import warnings
        warnings.warn("Use computePSDImages() instead", DeprecationWarning)
        self.computePSDImages(movie, fnUncorrected, fnCorrected,
                              outputFnUncorrected, outputFnCorrected)

    def computeThumbnail(self, inputFn, scaleFactor=6, outputFn=None):
        """ Generates a thumbnail of the input file"""
        outputFn = outputFn or self.getThumbnailFn(inputFn)
        args = "%s %s " % (inputFn, outputFn)
        args += "--meanshrink %s --fixintscaling=sane" % scaleFactor

        self.__runEman2Program('e2proc2d.py', args)

        return outputFn

    def correctGain(self, movieFn, outputFn, gainFn=None, darkFn=None):
        """correct a movie with both gain and dark images"""
        ih = emlib.image.ImageHandler()
        _, _, z, n = ih.getDimensions(movieFn)
        numberOfFrames = max(z, n)  # in case of wrong mrc stacks as volumes

        def _readImgFloat(fn):
            img = None
            if fn:
                img = ih.read(fn)
                img.convert2DataType(emlib.DT_FLOAT)
            return img

        gainImg = _readImgFloat(gainFn)
        darkImg = _readImgFloat(darkFn)

        img = ih.createImage()

        for i in range(1, numberOfFrames + 1):
            img.read((i, movieFn))
            img.convert2DataType(emlib.DT_FLOAT)

            if darkImg:
                img.inplaceSubtract(darkImg)
            if gainImg:
                img.inplaceMultiply(gainImg)

            img.write((i, outputFn))

    def getThumbnailFn(self, inputFn):
        """ Returns the default name for a thumbnail image"""
        return pwutils.replaceExt(inputFn, "thumb.png")

    def _getPsdCorr(self, movie):
        """ This should be implemented in subclasses."""
        pass

    def _writeFailedList(self, movieList):
        """ Write to a text file the items that have failed. """
        with open(self._getAllFailed(), 'a') as f:
            for movie in movieList:
                f.write('%d\n' % movie.getObjId())

    def _readFailedList(self):
        """ Read from a text file the id's of the items that have failed. """
        failedFile = self._getAllFailed()
        failedList = []
        if os.path.exists(failedFile):
            with open(failedFile) as f:
                failedList += [int(line.strip()) for line in f]

        return failedList


def createAlignmentPlot(meanX, meanY):
    """ Create a plotter with the cumulative shift per frame. """
    figureSize = (8, 6)
    plotter = Plotter(*figureSize)
    figure = plotter.getFigure()

    ax = figure.add_subplot(111)
    ax.grid()
    ax.axis('equal')
    ax.set_title('Cartesian representation')
    ax.set_xlabel('Drift x (pixels)')
    ax.set_ylabel('Drift y (pixels)')

    # Max range of the plot of the two coordinates
    plotRange = max(max(meanX) - min(meanX), max(meanY) - min(meanY))
    i = 1
    skipLabels = ceil(len(meanX) / 10.0)
    for x, y in izip(meanX, meanY):
        if i % skipLabels == 0:
            ax.text(x - 0.02 * plotRange, y + 0.02 * plotRange, str(i))
        i += 1

    ax.plot(meanX, meanY, color='b')
    ax.plot(meanX, meanY, 'yo')

    # setting the plot windows to properly see the data
    ax.axis([min(meanX) - 0.1 * plotRange, max(meanX) + 0.1 * plotRange,
             min(meanY) - 0.1 * plotRange, max(meanY) + 0.1 * plotRange])

    plotter.tightLayout()

    return plotter


class ProtAverageFrames(ProtAlignMovies):
    """
    Very simple protocol to align all the frames of a given data collection
    session. It can be used as a sanity check.
    """
    _label = 'average frames'

    # -------------------------- DEFINE param functions -----------------------
    def _defineAlignmentParams(self, form):
        pass

    def _processMovie(self, movie):
        allFramesSum = self._getPath('all_frames_sum.mrc')
        ih = emlib.image.ImageHandler()
        sumImg = ih.createImage()
        img = ih.createImage()

        n = movie.getNumberOfFrames()
        fn = movie.getFileName()

        sumImg.read((1, fn))

        for frame in range(2, n + 1):
            img.read((frame, fn))
            sumImg.inplaceAdd(img)

        if os.path.exists(allFramesSum):
            img.read(allFramesSum)
            sumImg.inplaceAdd(img)

        sumImg.write(allFramesSum)

    # FIXME: Methods will change when using the streaming for the output
    def createOutputStep(self):
        # Really load the input, since in the streaming case we can not
        # use the self.inputMovies directly
        allFramesSum = self._getPath('all_frames_sum.mrc')
        allFramesAvg = self._getPath('all_frames_avg.mrc')
        self._loadInputList()
        n = len(self.listOfMovies)

        ih = emlib.image.ImageHandler()
        sumImg = ih.read(allFramesSum)
        sumImg.inplaceDivide(float(n))
        sumImg.write(allFramesAvg)

        outputAvg = emobj.Image()
        outputAvg.setFileName(allFramesAvg)
        outputAvg.setSamplingRate(self.listOfMovies[0].getSamplingRate())
        self._defineOutputs(outputAverage=outputAvg)
        self._defineSourceRelation(self.inputMovies, outputAvg)

    def _validate(self):
        return []

    def _summary(self):
        return []

    def _createOutputMovies(self):
        """ Returns True if an output set of movies will be generated.
        The most common case is to always generate output movies,
        either with alignment only or the binary aligned movie files.
        Subclasses can override this function to change this behavior.
        """
        return False

    def _createOutputMicrographs(self):
        """ By default check if the user have selected 'doSaveAveMic'
        property. Subclasses can override this method to implement different
        behaviour.
        """
        return False

    def _createOutputWeightedMicrographs(self):
        return False



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es) [1]
# *              Kevin Savage (kevin.savage@diamond.ac.uk) [2]
# *
# * [1] Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# * [2] Diamond Light Source, Ltd
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import enum
import os
from os.path import join, basename
import re
from datetime import timedelta, datetime

import pyworkflow.utils as pwutils
import pyworkflow.protocol.params as params

from pwem import Domain
from pwem.emlib.image import ImageHandler
import pwem.constants as emcts

from .images import ProtImportImages
from ...objects import SetOfMicrographs, SetOfMovies


class ProtImportMicBase(ProtImportImages):
    """ Just to have a base class to both 
    ProtImportMicrographs and ProtImportMovies
    """
    _checkStacks = False

    def _defineAcquisitionParams(self, form):
        group = ProtImportImages._defineAcquisitionParams(self, form)

        group.addParam('samplingRateMode', params.EnumParam,
                       choices=[pwutils.Message.LABEL_SAMP_MODE_1,
                                pwutils.Message.LABEL_SAMP_MODE_2],
                       default=emcts.SAMPLING_FROM_IMAGE,
                       label=pwutils.Message.LABEL_SAMP_MODE,
                       help=pwutils.Message.TEXT_SAMP_MODE)

        group.addParam('samplingRate', params.FloatParam, default=1.0,
                       condition='samplingRateMode==%d' % emcts.SAMPLING_FROM_IMAGE,
                       label=pwutils.Message.LABEL_SAMP_RATE,
                       help=pwutils.Message.TEXT_SAMP_RATE)

        group.addParam('scannedPixelSize', params.FloatParam, default=7.0,
                       condition='samplingRateMode==%d' % emcts.SAMPLING_FROM_SCANNER,
                       label=pwutils.Message.LABEL_SCANNED,
                       help='')
        return group

    def _defineBlacklistParams(self, form):
        """ Options to blacklist certain items when launching the
        import protocol.
        """
        form.addSection(label="Rejection")
        form.addParam("blacklistSet", params.PointerParam,
                      pointerClass=self._getBlacklistSetClass(),
                      allowsNull=True,
                      label="Reject from Set",
                      help="Files on this set will not be imported")
        form.addParam('blacklistDateFrom', params.StringParam,
                      label="Reject from",
                      allowsNull=True,
                      help="Files acquired after this date will not be imported. "
                           "Must follow format: YYYY-mm-dd HH:MM:SS \n"
                           "e.g: 2019-01-14 14:18:05")
        form.addParam('blacklistDateTo', params.StringParam,
                      label="Reject before",
                      allowsNull=True,
                      help="Files acquired before this date will not be imported. "
                           "Must follow format: YYYY-mm-dd HH:MM:SS \n"
                           "e.g: 2019-01-14 14:18:05")
        form.addParam('useRegexps', params.BooleanParam,
                      default=True,
                      label='Rejection file has RegExps',
                      help="Choose Yes if the rejection file contains regular expressions. Set to No if "
                           "the rejection file contains file names. Ignore if not using a rejection file")
        form.addParam('blacklistFile', params.FileParam,
                      label="Blacklist File",
                      allowsNull=True,
                      help="Reject everything included in this file. If Use RegExps is True,"
                           "lines will be interpreted as regular expressions. E.g: \n"
                           "(.*)GRID_0[1-5](.*)\n"
                           "(.*)/GRID_10/Falcon_2019_01_14-16_(.*)\n"
                           "If Use RegExps is False, lines will be interpreted as file names. E.g.\n"
                           "/path/to/GRID_10/Falcon_2019_01_14-16_51_20_0_movie.mrcs\n"
                           "/path/to/GRID_10/Falcon_2019_01_14-16_55_40_0_movie.mrcs"
                      )

    def _getBlacklistSetClass(self):
        """ Returns the class to be blacklisted by this protocol.
        """
        return "SetOfImages"

    def _validateBlacklist(self):
        errors = []
        blacklistBySet = self.blacklistSet.get()
        if blacklistBySet and blacklistBySet.isStreamOpen():
            errors.append("Can't blacklist an open set. "
                          "Please stop streaming or  wait until streaming is done to blacklist this set.")

        dates = [self.blacklistDateFrom.get(), self.blacklistDateTo.get()]
        parsedDates = []
        for d in dates:
            if d:
                try:
                    parsedDates.append(datetime.strptime(d, "%Y-%m-%d %H:%M:%S"))
                except ValueError as e:
                    errors.append("Bad date formatting in blacklist date %s: %s" % (d, e))

        if len(parsedDates) == 2 and parsedDates[0] > parsedDates[1]:
            errors.append("Wrong blacklist dates: date from must be earlier than date to")
        return errors

    def _validate(self):
        errors = ProtImportImages._validate(self)

        errors += self._validateBlacklist()

        return errors

    def setSamplingRate(self, micSet):
        """ Set the sampling rate to the given set. """
        if self.samplingRateMode == emcts.SAMPLING_FROM_IMAGE:
            micSet.setSamplingRate(self.samplingRate.get())
        else:
            micSet.setScannedPixelSize(self.scannedPixelSize.get())

    def _acquisitionWizardCondition(self):
        """ By default this wizard will appears only when we import from
        a format that is not from files.
        But movie-import also can have a wizard to read from FEI xml files. """
        return 'True'

    def loadAcquisitionInfo(self):
        """ Return a proper acquisitionInfo (dict)
        or an error message (str).
        """
        if self.importFrom != self.IMPORT_FROM_FILES:
            return ProtImportImages.loadAcquisitionInfo(self)

        result = "Could not find acquisition information"

        for fileName, fileId in self.iterFiles():
            baseName = pwutils.removeExt(fileName)
            xml1 = baseName.replace('_frames', '.xml')
            if os.path.exists(xml1):
                result = self._parseXML(xml1)
            else:
                xml2 = baseName + '.xml'
                result = self._parseXML(xml2)

        return result

    def _parseXML(self, fileName):
        """ Parse micrograph XML files from FEI. """
        import xml.etree.ElementTree as ET

        # get context
        context = iter(ET.iterparse(fileName,
                                    events=('start', 'end')))

        labels = {'AccelerationVoltage': 'voltage',
                  'InstrumentModel': 'InstrumentModel',
                  'NominalMagnification': 'magnification'}

        # acq['amplitudeContrast'] = None
        # acq['sphericalAberration'] = None
        acq = {}

        def get(key, elem):
            acq[labels[key]] = elem.text

        pixelSize = False

        for event, elem in context:

            if event == 'start':
                if 'pixelSize' in elem.tag:
                    print("started: pixelSize")
                    pixelSize = True

            elif event == 'end':
                for l in labels:
                    if '}%s' % l in elem.tag:
                        get(l, elem)

                if '}numericValue' in elem.tag and pixelSize:
                    acq['samplingRate'] = float(elem.text) * 10e+09  # Convert to A
                    pixelSize = False

            else:
                raise Exception("Unknown event type %s" % event)

        # Correct for units conversion
        acq['voltage'] = float(acq['voltage']) / 1000.

        return acq

    def getItemsToBlacklistFromFile(self):
        if not hasattr(self, '_fileItemsToBlacklist'):
            blacklistfile = self.blacklistFile.get()
            blacklistItems = set()
            if blacklistfile:
                with open(blacklistfile, 'r') as f:
                    for blacklistedItem in f:
                        blacklistedItem = blacklistedItem.strip()
                        blacklistItems.add(blacklistedItem)
            self._fileItemsToBlacklist = blacklistItems

        return self._fileItemsToBlacklist

    def getBlacklistedItems(self):
        if not hasattr(self, '_blacklistedItems'):
            self._blacklistedItems = set()
        return self._blacklistedItems

    def isBlacklisted(self, fileName):
        # check if already blacklisted
        blacklistedItems = self.getBlacklistedItems()
        if fileName in blacklistedItems:
            return True

        # Blacklisted by set
        blacklistSet = self.blacklistSet.get()
        if blacklistSet is not None:
            for img in blacklistSet:
                blacklistFileName = img.getFileName()
                if ((os.path.islink(blacklistFileName)
                     and fileName == os.readlink(blacklistFileName))
                        or (self._getUniqueFileName(fileName) == os.path.basename(blacklistFileName))):
                    self.info("Blacklist warning: %s is blacklisted by the input set" % fileName)
                    blacklistedItems.add(fileName)
                    return True

        # Blacklisted by date
        blacklistDateFrom = self.blacklistDateFrom.get()
        blacklistDateTo = self.blacklistDateTo.get()
        doDateBlacklist = blacklistDateFrom is not None or blacklistDateTo is not None
        if doDateBlacklist:
            fileDate = datetime.fromtimestamp(os.path.getmtime(fileName))
            if blacklistDateFrom:
                parsedDateFrom = datetime.strptime(blacklistDateFrom, "%Y-%m-%d %H:%M:%S")
                if blacklistDateTo:
                    parsedDateTo = datetime.strptime(blacklistDateTo, "%Y-%m-%d %H:%M:%S")
                    if parsedDateFrom <= fileDate <= parsedDateTo:
                        self.info("Blacklist warning: %s is blacklisted by date" % fileName)
                        blacklistedItems.add(fileName)
                        return True
                else:
                    if parsedDateFrom <= fileDate:
                        self.info("Blacklist warning: %s is blacklisted by date" % fileName)
                        blacklistedItems.add(fileName)
                        return True

            elif blacklistDateTo:
                parsedDateTo = datetime.strptime(blacklistDateTo, "%Y-%m-%d %H:%M:%S")
                if fileDate <= parsedDateTo:
                    self.info("Blacklist warning: %s is blacklisted by date" % fileName)
                    blacklistedItems.add(fileName)
                    return True

        # Blacklisted by file
        items2blacklist = self.getItemsToBlacklistFromFile()
        for item2blacklist in items2blacklist:
            if self.useRegexps.get():
                if re.match(item2blacklist, fileName):
                    self.info("Blacklist warning: %s matched blacklist regexp %s"
                              % (fileName, item2blacklist))
                    blacklistedItems.add(fileName)
                    return True
            elif fileName in item2blacklist:
                self.info("Blacklist warning: %s is blacklisted " % fileName)
                blacklistedItems.add(fileName)
                return True
        return False

class ImportMicsOutput(enum.Enum):
    outputMicrographs = SetOfMicrographs

class ProtImportMicrographs(ProtImportMicBase):
    """Protocol to import a set of micrographs to the project"""
    _label = 'import micrographs'
    _possibleOutputs = ImportMicsOutput
    _outputClassName = ImportMicsOutput.outputMicrographs.value.__name__

    IMPORT_FROM_EMX = 1
    IMPORT_FROM_XMIPP3 = 2
    IMPORT_FROM_SCIPION = 3

    def _getImportChoices(self):
        """ Return a list of possible choices
        from which the import can be done.
        (usually packages formats such as: xmipp3, eman2, relion...etc.
        """
        choices = ProtImportImages._getImportChoices(self)
        return choices + ['emx', 'xmipp3', 'scipion']

    def _getBlacklistSetClass(self):
        """ Returns the class to be blacklisted by this protocol.
        """
        return "SetOfMicrographs"

    def _defineImportParams(self, form):
        """ Just redefine to put some import parameters
        before the acquisition related parameters.
        """
        form.addParam('emxFile', params.FileParam,
                      condition='(importFrom == %d)' % self.IMPORT_FROM_EMX,
                      label='Input EMX file',
                      help="Select the EMX file containing micrographs information.\n"
                           "See more about [[http://i2pc.cnb.csic.es/emx][EMX format]]")

        form.addParam('mdFile', params.FileParam,
                      condition='(importFrom == %d)' % self.IMPORT_FROM_XMIPP3,
                      label='Micrographs metadata file',
                      help="Select the micrographs Xmipp metadata file.\n"
                           "It is usually a _micrograph.xmd_ file result\n"
                           "from import, preprocess or downsample protocols.")

        form.addParam('sqliteFile', params.FileParam,
                      condition='(importFrom == %d)' % self.IMPORT_FROM_SCIPION,
                      label='Micrographs sqlite file',
                      help="Select the micrographs sqlite file.\n")

    # --------------------------- INSERT functions ----------------------------
    def _insertAllSteps(self):
        importFrom = self.importFrom.get()
        ci = self.getImportClass()

        if ci is None:
            ProtImportMicBase._insertAllSteps(self)
        else:
            self._insertFunctionStep('importMicrographsStep', importFrom,
                                     self.importFilePath)

    # --------------------------- STEPS functions -----------------------------
    def importMicrographsStep(self, importFrom, *args):
        ci = self.getImportClass()
        ci.importMicrographs()

        summary = "Import from *%s* file:\n" % self.getEnumText('importFrom')
        summary += self.importFilePath + '\n'

        if self.hasAttribute('outputParticles'):
            particles = self.outputParticles
            summary += '   Particles: *%d* (ctf=%s, alignment=%s)\n' % (particles.getSize(),
                                                                        particles.hasCTF(),
                                                                        particles.getAlignment())

        if self.hasAttribute('outputCoordinates'):  # EMX files can contain only Coordinates information
            summary += '   Coordinates: *%d* \n' % (self.outputCoordinates.getSize())

        if self.hasAttribute('outputMicrographs'):  # EMX files can contain only Coordinates information
            summary += '   Micrographs: *%d* \n' % (self.outputMicrographs.getSize())

        if self.copyFiles:
            summary += ('\n_WARNING_: Binary files copied into project '
                        '(extra disk space)')

        self.summaryVar.set(summary)

    # --------------------------- INFO functions ------------------------------
    def _validate(self):
        ci = self.getImportClass()
        if ci is None:
            errors = ProtImportMicBase._validate(self)
            for micFn, _ in self.iterFiles():
                imgh = ImageHandler()
                if imgh.isImageFile(micFn):
                    _, _, z, n = imgh.getDimensions(micFn)
                    if n > 1 or z > 1:
                        errors.append("The protocol not support micrographs "
                                      "stored in stacks. If you want to "
                                      "obtain your micrographs individually, "
                                      "you can run the following command:\n"
                                      "scipion run scipion_directory/scripts/"
                                      "split_stacks.py --files *your files* "
                                      "--ext *extension*")
                # JMRT: only check the first image, for large dataset
                # even reading the header can take a while
                break
            return errors

        else:
            return ci.validateMicrographs()

    def _summary(self):
        if self.importFrom == self.IMPORT_FROM_FILES:
            return ProtImportMicBase._summary(self)
        else:
            return [self.summaryVar.get('No summary information.')]

    # --------------------------- UTILS functions -----------------------------
    def getImportClass(self):
        """ Return the class in charge of importing the files. """
        if self.importFrom == self.IMPORT_FROM_EMX:
            EmxImport = Domain.importFromPlugin('emxlib.convert', 'EmxImport')
            self.importFilePath = self.emxFile.get('').strip()
            return EmxImport(self, self.importFilePath)
        elif self.importFrom == self.IMPORT_FROM_XMIPP3:
            XmippImport = Domain.importFromPlugin('xmipp3.convert', 'XmippImport')
            self.importFilePath = self.mdFile.get('').strip()
            return XmippImport(self, self.mdFile.get())
        elif self.importFrom == self.IMPORT_FROM_SCIPION:
            from .dataimport import ScipionImport
            self.importFilePath = self.sqliteFile.get('').strip()
            return ScipionImport(self, self.importFilePath)
        else:
            self.importFilePath = ''
            return None

class ImportMoviesOutput(enum.Enum):
    outputMovies = SetOfMovies

class ProtImportMovies(ProtImportMicBase):
    """ Protocol to import a set of movies (from direct detector cameras)
    to the project.
    """
    _label = 'import movies'
    _possibleOutputs = ImportMoviesOutput
    _outputClassName = ImportMoviesOutput.outputMovies.value.__name__

    def __init__(self, **kwargs):
        ProtImportMicBase.__init__(self, **kwargs)
        self.serverSocket = None
        self.connectionList = None

    def _getBlacklistSetClass(self):
        """ Returns the class to be blacklisted by this protocol.
        """
        return "SetOfMovies"

    def _defineAcquisitionParams(self, form):
        group = ProtImportMicBase._defineAcquisitionParams(self, form)

        line = group.addLine('Dose (e/A^2)',
                             help="Initial accumulated dose (usually 0) and "
                                  "dose per frame. ")

        line.addParam('doseInitial', params.FloatParam, default=0,
                      label='Initial')

        line.addParam('dosePerFrame', params.FloatParam, default=None,
                      allowsNull=True,
                      label='Per frame')

        form.addParam('gainFile', params.FileParam,
                      label='Gain image',
                      help='A gain reference related to a set of movies'
                           ' for gain correction')

        form.addParam('darkFile', params.FileParam,
                      label='Dark image',
                      help='A dark image related to a set of movies')

    def _defineParams(self, form):
        ProtImportMicBase._defineParams(self, form)

        form.addSection('Frames')

        streamingConditioned = "dataStreaming"
        framesCondition = "inputIndividualFrames"

        form.addParam('inputIndividualFrames', params.BooleanParam,
                      default=False,
                      label="Input individual frames?",
                      help="Select Yes if movies are acquired in individual "
                           "frame files. ")
        form.addParam('numberOfIndividualFrames', params.IntParam,
                      condition=framesCondition,
                      label='Number of frames',
                      help='Provide how many frames are per movie. ')
        form.addParam('stackFrames', params.BooleanParam,
                      default=False, condition=framesCondition,
                      label="Create movie stacks?",
                      help="Select Yes if you want to create a new stack for "
                           "each movies with its frames. ")
        # This is not working so for now its hidden
        form.addParam('writeMoviesInProject', params.BooleanParam,
                      default=False,
                      condition=framesCondition + " and stackFrames",
                      label="Write stacks in the project folder?",
                      help="If Yes, the created stack files will be written "
                           "in the project folder. By default the movies will "
                           "be written in the same place where input frames "
                           "are.")
        form.addParam('movieSuffix', params.StringParam,
                      default='_frames.mrcs',
                      condition=framesCondition + " and stackFrames",
                      label="Movie suffix",
                      help="Suffix added to the output movie filename."
                           "Use the extension to select the format ("
                           "e.g., .mrcs, .stk)")
        form.addParam('deleteFrames', params.BooleanParam,
                      default=False,
                      condition=framesCondition + " and stackFrames",
                      label="Delete frame files?",
                      help="Select Yes if you want to remove the individual "
                           "frame files after creating the movie stack. ")

        streamingSection = form.getSection('Streaming')
        streamingSection.addParam('moviesToExclude', params.PointerParam,
                                  pointerClass='SetOfMovies',
                                  condition=streamingConditioned,
                                  allowsNull=True,
                                  expertLevel=params.LEVEL_ADVANCED,
                                  label="Previous movies to exclude",
                                  help="Select a setOfMovies that are already "
                                       "imported that you want to exclude for "
                                       "this import.")

    # --------------------------- INSERT functions ----------------------------
    def _insertAllSteps(self):
        # Only the import movies has property 'inputIndividualFrames'
        # so let's query in a non-intrusive manner
        inputIndividualFrames = getattr(self, 'inputIndividualFrames', False)

        if self.dataStreaming or inputIndividualFrames:
            funcName = 'importImagesStreamStep'
        else:
            funcName = 'importImagesStep'

        self._insertFunctionStep(funcName, self.getPattern(),
                                 self.voltage.get(),
                                 self.sphericalAberration.get(),
                                 self.amplitudeContrast.get(),
                                 self.magnification.get())

    # --------------------------- INFO functions -------------------------------
    def _validate(self):
        """Overwriting to skip file validation if streaming with socket"""

        errors = ProtImportMicBase._validate(self)
        if self.inputIndividualFrames and not self.stackFrames:
            errors.append("Scipion does not support individual frames. "
                          "You must set to Yes *Create movie stacks?* "
                          "parameter.")

        if not self.gainFile.empty() and not os.path.exists(self.gainFile.get()):
            errors.append("Gain file not found in " + str(self.gainFile.get()))
        if not self.darkFile.empty() and not os.path.exists(self.darkFile.get()):
            errors.append("Dark file not found in " + str(self.gainFile.get()))
        return errors

    # --------------------------- UTILS functions ------------------------------
    def setSamplingRate(self, movieSet):
        ProtImportMicBase.setSamplingRate(self, movieSet)
        movieSet.setGain(self.gainFile.get())
        movieSet.setDark(self.darkFile.get())
        acq = movieSet.getAcquisition()
        acq.setDoseInitial(self.doseInitial.get())
        acq.setDosePerFrame(self.dosePerFrame.get())

    def _setupFirstImage(self, movie, imgSet):
        # Create a movie object to read dimensions
        dimMovie = movie.clone()
        movieFn = movie.getFileName()

        def decompress(program, args, ext, nExt):
            movieFolder = self._getTmpPath()
            movieName = basename(movie.getFileName())
            movieTmpLink = join(movieFolder, movieName)
            pwutils.cleanPath(movieTmpLink)
            pwutils.createAbsLink(os.path.abspath(movieFn), movieTmpLink)
            self.runJob(program, args % movieName, cwd=movieFolder)
            dimMovie.setFileName(movieTmpLink.replace(ext, nExt))

        if movieFn.endswith('bz2'):
            decompress('bzip2', '-d -f %s', '.bz2', '')

        elif movieFn.endswith('tbz'):
            decompress('tar', 'jxf %s', '.tbz', '.mrc')

        dim = dimMovie.getDim()
        self.info("Dim: (%s)" % ", ".join(map(str, dim)))
        range = [1, dim[2], 1]
        movie.setFramesRange(range)
        imgSet.setDim(dim)
        imgSet.setFramesRange(range)

    def iterNewInputFiles(self):
        """ In the case of importing movies, we want to override this method
        for the case when input are individual frames and we want to create
        movie stacks before importing.
        The frames pattern should contains a part delimited by $.
        The id expression with # is not supported for simplicity.
        """

        if not (self.inputIndividualFrames and self.stackFrames):
            # In this case behave just as
            iterInputFiles = ProtImportMicBase.iterNewInputFiles(self)

            for fileName, uniqueFn, fileId in iterInputFiles:
                yield fileName, uniqueFn, fileId
            return

        if self.dataStreaming:
            # Consider only the files that are not changed in the fileTime
            # delta if processing data in streaming
            fileTimeout = timedelta(seconds=self.fileTimeout.get())
            filePaths = [f for f in self.getMatchFiles()
                         if not self.fileModified(f, fileTimeout)]
        else:
            filePaths = self.getMatchFiles()

        frameRegex = re.compile("(?P<prefix>.+[^\d]+)(?P<frameid>\d+)")
        # Group all frames for each movie
        # Key of the dictionary will be the common prefix and the value
        # will be a list with all frames in that movie
        frameDict = {}

        for fileName in filePaths:
            fnNoExt = pwutils.removeExt(fileName)

            match = frameRegex.match(fnNoExt)

            if match is None:
                raise Exception("Incorrect match of frame files pattern!")

            d = match.groupdict()
            prefix = d['prefix']
            frameid = int(d['frameid'])

            if prefix not in frameDict:
                frameDict[prefix] = []

            frameDict[prefix].append((frameid, fileName))

        suffix = self.movieSuffix.get()
        ih = ImageHandler()

        for movieFn in self.createdStacks:
            uniqueFn = basename(movieFn)
            if uniqueFn not in self.importedFiles:
                yield movieFn, uniqueFn, None

        def checkMovie():
            for k, v in frameDict.items():
                moviePath = os.path.dirname(k)
                movieFn = join(moviePath + "/", self._getUniqueFileName(k) +
                               suffix)

                if self.writeMoviesInProject:
                    movieFn = self._getExtraPath(os.path.basename(movieFn))

                if (movieFn not in self.importedFiles and
                        movieFn not in self.createdStacks and
                        len(v) == self.numberOfIndividualFrames):
                    movieOut = movieFn

                    if movieOut.endswith("mrc"):
                        movieOut += ":mrcs"

                    # By default we will write the movie stacks
                    # unless we are in continue mode and the file exists
                    writeMovie = True
                    if self.isContinued() and os.path.exists(movieFn):
                        self.info("Skipping movie stack: %s, seems to be done"
                                  % movieFn)
                        writeMovie = False

                    if writeMovie:
                        self.info("Writing movie stack: %s" % movieFn)
                        # Remove the output file if exists
                        pwutils.cleanPath(movieFn)

                        for i, frame in enumerate(sorted(v, key=lambda x: x[0])):
                            frameFn = frame[1]  # Frame name stored previously
                            ih.convert(frameFn, (i + 1, movieOut))

                            if self.deleteFrames:
                                pwutils.cleanPath(frameFn)

                    # Now return the newly created movie file as imported file
                    self.createdStacks.add(movieFn)
                    return

        checkMovie()

    def ignoreCopy(self, source, dest):
        pass

    def getCopyOrLink(self):
        if (self.inputIndividualFrames and self.stackFrames and
                self.writeMoviesInProject):
            return self.ignoreCopy
        else:
            return ProtImportMicBase.getCopyOrLink(self)



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This modules contains constants related to EM
"""
# ------------------ Constants values --------------------------------------

NO_INDEX = 0  # This is the index value for single images
URL = 'https://github.com/scipion-em/scipion-em'
# Sampling rate input modes
SAMPLING_FROM_IMAGE = 0
SAMPLING_FROM_SCANNER = 1

RELATION_CTF = 'relation_ctf'

UNIT_PIXEL = 'px'
UNIT_PIXEL_FOURIER = '1/px'
UNIT_ANGSTROM = 'A'
UNIT_ANGSTROM_FOURIER = '1/A'
# No needed: Å should be valid as long as first line
# in file has utf-8 directive --> # -*- coding: utf-8 -*-
UNIT_ANGSTROM_SYMBOL = u"\u212B"

# Fourier Filter options
FILTER_LOW_PASS = 0
FILTER_HIGH_PASS = 1
FILTER_BAND_PASS = 2
FILTER_GAUSSIAN = 3
FILTER_LOW_PASS_NO_DECAY = 4
FILTER_NO_DECAY = 5

# Transform
ALIGN_NONE = 'None'
ALIGN_2D = '2D'            # 2D image alignment
ALIGN_3D = '3D'            # 3D map alignment
ALIGN_PROJ = 'Projection'  # relate projections with 3d map

ALIGNMENTS = [ALIGN_NONE, ALIGN_2D, ALIGN_3D, ALIGN_PROJ]

# Constants related with colormaps for viewers
# Color maps
COLOR_JET = 0
COLOR_TERRAIN = 1
COLOR_GIST_EARTH = 2
COLOR_GIST_NCAR = 3
COLOR_GNU_PLOT = 4
COLOR_GNU_PLOT2 = 5
COLOR_OTHER = 6

COLOR_CHOICES = ['jet', 'terrain', 'gist_earth',
                 'gist_ncar', 'gnuplot', 'gnuplot2', 'other']

# Axis code
AX_X = 0
AX_Y = 1
AX_Z = 2

# SYMMETRY, conventions described at:
# https://scipion-em.github.io/docs/docs/developer/symmetries
# Some notes:
# Icosahedral
#   xmipp and relion define I1, I2, I3 and I4 that corerspond to
#   I222, I222r, In25 and In25z
#   cryosparc has I1 and I2
#   EMAN always puts the highest symmetry axis on Z.
#   and a 2 fold axis in X (so it uses I2n5 or I2n5r, not sure)
# DIEDRAL: first axis Z, second X (DX) except cryosparc that uses y (DY)
# Tetraedral, most of the programs use TZ3
# CN cyclic symmetry Cn around z axis
SYM_CYCLIC = 0
# DN cyclic symmetry plus and extra 2fold symmetry axis around X
SYM_DIHEDRAL = 1
SYM_DIHEDRAL_X = SYM_DIHEDRAL
# DN cyclic symmetry plus and extra 2fold symmetry axis around Y (cryoSparc)
SYM_DIHEDRAL_Y = 2
# T 222 Tetrahedral symmetry with two-fold symmetry axis along the X, Y and Z
SYM_TETRAHEDRAL = 3
SYM_TETRAHEDRAL_222 = SYM_TETRAHEDRAL
# T  three-fold symmetry axis along Z, another three-fold axis
# in the YZ plane such that the sign(y) != sign(z)
SYM_TETRAHEDRAL_Z3 = 4
# SYM_TETRAHEDRAL222R  is a degenerated symmetry, that is
# for each symmetry axis (x,y,z)  defined in T222, there is a
# symmetry axis (-x,-y,-z) in T222R, therefore the produced
# symmetry matrices are equivalent
# SYM_TETRAHEDRAL222R = 14

# T  three-fold symmetry axis along Z, another three-fold axis
# in the YZ plane such that  sign(y) = sign(z)
SYM_TETRAHEDRAL_Z3R = 5

SYM_OCTAHEDRAL = 6  # O

# icosahedric IXXX
# (no crowther 222 and standard in heyman et al 2005 article).
# 2-fold axes on x,y,z axes. With the positive z-axis pointing at the viewer,
# the front-most 5-fold vertices are in yz plane, and the front-most 3-fold
# axes are in the xz plane.
SYM_I222 = 7

# (crowther) 2-fold axes on x,y,z axes. With the positive z-axis pointing at
# the viewer, the front-most 5-fold vertices are in xz plane,
# and the front-most 3-fold axes are in the yz plane.
# (222 rotated 90 degrees around Z)
SYM_I222r = 8

# '2-fold symmetry along y and 5-fold along z
SYM_In25 = 9

# 'n25' with 180 degree rotation about x
SYM_In25r = 10
SYM_I2n3 = 11  # Two-fold symmetry along X and 3-fold along Z
SYM_I2n3r = 12  # Idem but rotated 180 degree about Y
SYM_I2n5 = 13  # Two-fold symmetry along Y and 5-fold along Z
SYM_I2n5r = 14  # Idem but rotated 180 degree about X

# Symmetry dictionary
SCIPION_SYM_NAME = dict()
SCIPION_SYM_NAME[SYM_CYCLIC] = 'Cn'
SCIPION_SYM_NAME[SYM_DIHEDRAL_X] = 'Dxn'
SCIPION_SYM_NAME[SYM_DIHEDRAL_Y] = 'Dyn'
SCIPION_SYM_NAME[SYM_TETRAHEDRAL] = 'T222'
SCIPION_SYM_NAME[SYM_TETRAHEDRAL_Z3] = 'Tz3'
SCIPION_SYM_NAME[SYM_TETRAHEDRAL_Z3R] = 'Tz3r'
SCIPION_SYM_NAME[SYM_OCTAHEDRAL] = 'O'
SCIPION_SYM_NAME[SYM_I222] = 'I222'
SCIPION_SYM_NAME[SYM_I222r] = 'I222r'
SCIPION_SYM_NAME[SYM_In25] = 'In25'
SCIPION_SYM_NAME[SYM_In25r] = 'In25r'
SCIPION_SYM_NAME[SYM_I2n3] = 'I2n3'
SCIPION_SYM_NAME[SYM_I2n3r] = 'I2n3r'
SCIPION_SYM_NAME[SYM_I2n5] = 'I2n5'
SCIPION_SYM_NAME[SYM_I2n5r] = 'I2n5r'

# Entry points names
CONVERT_ENTRY_POINT = 'emconvert'
CHIMERA_ENTRY_POINT = 'emchimera'
EM_PROGRAM_ENTRY_POINT = 'emprogram'

# maxit
MAXIT_HOME = 'MAXIT_HOME'
MAXIT = 'maxit'

# Explanation messages for the wizards
RING_MASK_WIZ_MSG = \
    'The values of the outer and inner radius can be controlled via '\
    'both the sliders placed below or '\
    'the mousewheel (when the mouse cursor is over the image). '\
    'The arrow (=>) located at the left of '\
    'the sliders is used to indicate which one is active. '\
    'To change from one to another, simply '\
    'click and drag the corresponding slider or use up and down keys.'
CIRCLE_MASK_WIZ_MSG = \
    'The values of the mask radius can be controlled via '\
    'both the slider or the mousewheel (when ' \
    'the mouse cursor is over the image).'
FREQ_BANDPASS_WIZ_MSG =\
    'The values of the low and high frequencies can '\
    'be controlled via both the sliders placed below or '\
    'the mousewheel (when the mouse cursor is over the image). '\
    'The arrow (=>) located at the left of '\
    'the sliders is used to indicate which one is active. '\
    'To change from one to another, simply '\
    'click and drag the corresponding slider or use up and down keys. '\
    'Values are shown in both digital '\
    'frequency and angstroms.'

EM_ROOT_VAR = 'EM_ROOT'

DEFAULT_MAX_PREVIEW_FILE_SIZE = 500  # Unit is MB: 500 MB

RESIDUES3TO1 = {'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K',
                'ILE': 'I', 'PRO': 'P', 'THR': 'T', 'PHE': 'F', 'ASN': 'N',
                'GLY': 'G', 'HIS': 'H', 'LEU': 'L', 'ARG': 'R', 'TRP': 'W',
                'ALA': 'A', 'VAL': 'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M'}

RESIDUES1TO3 = {v: k for k, v in RESIDUES3TO1.items()}


# File extensions
# MRC
EXT_MRC ="mrc"
EXT_MRCS ="mrcs"
EXT_MRC_MRC="%s:%s" % (EXT_MRC,EXT_MRC)
EXT_MRC_MRCS="%s:%s" % (EXT_MRC,EXT_MRCS)
EXT_MRCS_MRC="%s:%s" % (EXT_MRCS,EXT_MRC)
EXT_MRCS_MRCS="%s:%s" % (EXT_MRCS,EXT_MRCS)
EXT_ST ="st"
EXT_REC = "rec"
EXT_ALI = "ali"
EXT_MAP = "map"

ALL_MRC_EXTENSIONS =[EXT_MRC, EXT_MRCS, EXT_MRC_MRC, EXT_MRC_MRCS, EXT_MRCS_MRC, EXT_MRCS_MRCS,
                     EXT_ST, EXT_REC, EXT_ALI, EXT_MAP]

ALL_TIF_EXTENSIONS = ['tif','tiff', 'gain', 'eer']



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os

import pyworkflow.viewer as pwviewer
import pyworkflow.utils as pwutils
import pyworkflow.object as pwobj

# FIXME: rename data by objects
import pwem.objects as emobj
import pwem.emlib.metadata as md
from pwem.viewers import ObjectView


class EmProtocolViewer(pwviewer.ProtocolViewer):
    """ Subclass of ProtocolViewer that have specific functions
    for EM data visualization.
    """
    def objectView(self, filenameOrObject, **kwargs):
        """ This is a wrapper around the ObjectView constructor, just to
        avoid passing the project and protocol, since both are know
        here in the ProtocolViewer.
        Params:
            filenameOrObject: This parameter can be either a filename or an
                object that has 'getFileName' method.
            **kwargs: Can receive extra keyword-arguments that will be passed
                to the ObjectView constructor
        """
        fn = None

        if isinstance(filenameOrObject, str):
            # If the input is a string filename, we should take the object id
            # from the protocol. This assumes that self.protocol have been
            # previously set
            fn = filenameOrObject
            strId = self.getProtocolId()

        elif isinstance(filenameOrObject, pwobj.Object):

            if filenameOrObject.getObjId() is None:
                strId = self.getProtocolId()
            else:
                strId = filenameOrObject.strId()

            if hasattr(filenameOrObject, 'getLocation'):
                # In this case fn will be a location tuple that will be
                # correctly handled by the showj DataView
                fn = filenameOrObject.getLocation()
            elif hasattr(filenameOrObject, 'getFileName'):
                # If the input is an object, we can take the id from it
                fn = filenameOrObject.getFileName()

        if fn is None:
            raise Exception("Incorrect input object, it should be 'string' or "
                            "'Object' (with 'getLocation' or 'getFileName' "
                            "methods).")

        return ObjectView(self._project, strId, fn, **kwargs)

    def createVolumesSqlite(self, files, path, samplingRate,
                            updateItemCallback=None):
        pwutils.cleanPath(path)
        volSet = emobj.SetOfVolumes(filename=path)
        volSet.setSamplingRate(samplingRate)

        for volFn in files:
            vol = emobj.Volume()
            vol.setFileName(volFn)
            if updateItemCallback:
                updateItemCallback(vol)
            volSet.append(vol)
        volSet.write()
        volSet.close()

        return volSet

    def createAngDistributionSqlite(self, sqliteFn, numberOfParticles,
                                    itemDataIterator):
        if not os.path.exists(sqliteFn):
            # List of list of 3 elements containing angleTilt, anglePsi, weight
            projectionList = []

            def getCloseProjection(angleRot, angleTilt):
                """ Get an existing projection close to angleRot, angleTilt.
                Return None if not found close enough.
                """
                for projection in projectionList:
                    if (abs(projection[0] - angleRot) <= 0.5 and
                            abs(projection[1] - angleTilt) <= 0.5):
                        return projection
                return None

            weight = 1. / numberOfParticles

            for angleRot, angleTilt in itemDataIterator:
                projection = getCloseProjection(angleRot, angleTilt)
                if projection is None:
                    projectionList.append([angleRot, angleTilt, weight])
                else:
                    projection[2] = projection[2] + weight

            mdProj = md.MetaData()

            for projection in projectionList:
                mdRow = md.Row()
                mdRow.setValue(md.MDL_ANGLE_ROT, projection[0])
                mdRow.setValue(md.MDL_ANGLE_TILT, projection[1])
                mdRow.setValue(md.MDL_WEIGHT, projection[2])
                mdRow.writeToMd(mdProj, mdProj.addObject())
            mdProj.write(sqliteFn)



# **************************************************************************
# *
# * Authors:     Roberto Marabini
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module returns the matrices related with the different
point symmetries. Code based on chimera file sym.py
"""

import logging
logger = logging.getLogger(__name__)
from math import sin, cos, pi, acos, sqrt, asin
import numpy as np
from numpy.linalg import inv as matrix_inverse
from numpy import dot as matrix_multiply
import operator

import pwem.constants as cts

DEBUG = False  # set to True for debuging
if DEBUG:
    from pwem.viewers.viewer_chimera import Chimera
    # expansion factor is use to create chimera bild files
    # this is the factor to multiply the coordinates
    expansionFactor = 100


def _column(matrix, i):
    """Returns column i of matrix as a list."""
    return [row[i] for row in matrix]


def _applyMatrix(tf, points):
    """
    Args:multiply point by a matrice list
    """
    tf = np.array(tf)
    r = matrix_multiply(points, np.transpose(tf[:, :3]))
    np.add(r, tf[:, 3], r)
    return r


def __length(v):
    """Returns the length of a vector."""
    d = sqrt(sum([e*e for e in v]))
    return d


def _normalizeVector(v):
    """Returns a unit vector in the same direction as v."""
    d = __length(v)
    if d == 0:
        d = 1
    return tuple([e/d for e in v])


def _reflectMatrix(axis):
    """Returns a matrix that reflects points in the plane
    perpendicular to axis."""
    axis = _normalizeVector(axis)
    return ((1 - 2*axis[0]*axis[0], -2*axis[0]*axis[1], -2*axis[0]*axis[2], 0),
            (-2*axis[1]*axis[0], 1 - 2*axis[1]*axis[1], -2*axis[1]*axis[2], 0),
            (-2*axis[2]*axis[0], -2*axis[2]*axis[1], 1 - 2*axis[2]*axis[2], 0))


def _rotationTransform(axis, angle, center=(0, 0, 0)):
    """ returns matrix that rotates "angle" angles arround vector axis
        angle is in degrees
    """
    axis = _normalizeVector(axis)

    arad = angle*pi/180.0
    sa = sin(arad)
    ca = cos(arad)
    if axis == (0, 0, 1) and center == (0, 0, 0):
        return ((ca, -sa, 0, 0),
                (sa, ca, 0, 0),
                (0, 0, 1, 0))

    k = 1 - ca
    ax, ay, az = axis
    tf = ((1 + k*(ax*ax-1), -az*sa+k*ax*ay, ay*sa+k*ax*az, 0),
          (az*sa+k*ax*ay, 1 + k*(ay*ay-1), -ax*sa+k*ay*az, 0),
          (-ay*sa+k*ax*az, ax*sa+k*ay*az, 1 + k*(az*az-1), 0))

    if center == (0, 0, 0):
        return tf

    cx, cy, cz = center
    c_tf = ((1, 0, 0, cx), (0, 1, 0, cy), (0, 0, 1, cz))
    inv_c_tf = ((1, 0, 0, -cx), (0, 1, 0, -cy), (0, 0, 1, -cz))
    rtf = _multiplyMatrices(c_tf, tf, inv_c_tf)
    return rtf


def _translationMatrix(shift):
    """ returns matriz that shift point by vector shift"""
    tf = np.array(((1.0, 0, 0, shift[0]),
                   (0, 1.0, 0, shift[1]),
                   (0, 0, 1.0, shift[2])))
    return tf


def _identityMatrix():
    """ returns identity matrix"""
    return (1.0, 0, 0, 0), (0, 1.0, 0, 0), (0, 0, 1.0, 0)


def _invertMatrix(tf):
    """ returns inverse 3x3 matrix"""
    tf = np.array(tf)
    r = tf[:, :3]
    t = tf[:, 3]
    tfinv = np.zeros((3, 4), np.float)
    rinv = tfinv[:, :3]
    tinv = tfinv[:, 3]
    rinv[:, :] = matrix_inverse(r)
    tinv[:] = matrix_multiply(rinv, -t)
    return tfinv


def _multiplyMatrices(*mlist):
    """ returns matrix product of matrices in mlist"""
    if len(mlist) == 2:
        m1, m2 = mlist
        p = [[0, 0, 0, 0],
             [0, 0, 0, 0],
             [0, 0, 0, 0]]
        for r in range(3):
            for c in range(4):
                p[r][c] = m1[r][0]*m2[0][c] +\
                          m1[r][1]*m2[1][c] +\
                          m1[r][2]*m2[2][c]
            p[r][3] += m1[r][3]
        p = tuple(map(tuple, p))
    else:
        p = _multiplyMatrices(*mlist[1:])
        p = _multiplyMatrices(mlist[0], p)
    return p


def _matrixProducts(mlist1, mlist2):
    """ returns list of matrix products of matrices in mlist1 and mlist2"""
    plist = []
    for m1 in mlist1:
        for m2 in mlist2:
            m1xm2 = _multiplyMatrices(m1, m2)
            plist.append(m1xm2)
    return plist


def _coordinateTransformList(tflist, ctf):
    """ returns list of coordinate transforms of tflist by ctf"""
    ctfinv = _invertMatrix(ctf)
    return [_multiplyMatrices(ctfinv, tf, ctf) for tf in tflist]


def _recenterSymmetries(tflist, center):

    if center == (0, 0, 0):
        return tflist
    ctf = _translationMatrix([-x for x in center])
    return _coordinateTransformList(tflist, ctf)


def _transposeMatrix(tf):
    """ returns transpose of 3x3 matrix"""
    return ((tf[0][0], tf[1][0], tf[2][0], tf[0][3]),
            (tf[0][1], tf[1][1], tf[2][1], tf[1][3]),
            (tf[0][2], tf[1][2], tf[2][2], tf[2][3]))

# ==================== End of utility functions ====================

class SymmetryHelper:
    """ This is a static class so there is no need to instantiate it. 
        Given a symmetry the class provides methods to "move" projection
        directions to the equivalent orientation that lays within the unit 
        cells defined at https://scipion-em.github.io/docs/release-3.0.0/docs/developer/symmetries/symmetries.html?highlight=symmetry."""

    # Dictionary to cache all the symmetry matrices and unit cell 
    # planes for a particular symmetry
    _matricesAndPlanes = {}  

    @classmethod
    def getSymmetryKey(cls,sym, n):
        return "%s_%s" % (sym, n)

    @classmethod
    def getSymmetryMatricesAndPlanes(cls, sym, n):
        """ Lazy return a tuple of symmetry matrices and unit cell planes for the symmetry and order passed

        :param sym: Symmetry type
        :param n: symmetry order

        """

        key = cls.getSymmetryKey(sym, n)
        if key not in cls._matricesAndPlanes.keys():

            logger.info("Gettting symmetry matrices and unitcell planes for %s, %s." % (sym,n))
            matrixSet = getSymmetryMatrices(sym=sym, n=n)
            # get unit cell vectors
            _, unitCellPlanes = getUnitCell(sym=sym, n=n, generalize=False)

            cls._matricesAndPlanes[key] = (matrixSet, unitCellPlanes)

        return cls._matricesAndPlanes[key]

    @classmethod
    def moveParticleInsideUnitCell(cls, particle, symmetry, symmetryOrder):
        """ apply to the projection direction the symmetry matrix that "moves" the projection
        direction to the equivalent orientation that lays within the "canonical" unit
        This is useful to move for example Cryosparc angles to a same unit cell
        and therefore been able to plot angular distribution. This method can be used in the typical "updateParticle"
        callback when generating any output set.

        :param particle: Particle to move.
        :param symmetry: symmetry type in scipion's convention.
        :param symmetryOrder: order of the symmetry

        """

        # For C1 particles we do not move angles to unit cell.
        if symmetry==cts.SYM_CYCLIC and symmetryOrder==1:
            logger.debug("Cancelling unit cell migration due to C1 symmetry.")
            return

        matrices, unitCellPlanes = cls.getSymmetryMatricesAndPlanes(symmetry, symmetryOrder)
        cls._moveParticleInsideUnitCell(particle, matrices, unitCellPlanes)


    @classmethod
    def _moveParticleInsideUnitCell(cls, particle, matrixSet, unitCellPlanes):
        """ Move the particle projection direction so it lays inside the canonical unit cell,
            symmetry matrices and unit cell planes needs to be precomputed
            :param particle: to move
            :param matrixSet: value returned when calling getSymmetryMatrices with the right symmetry and order
            :param unitCellPlanes: second term returned by getUnitCell when called with the right symmetry and order
            """
        # get projection direction
        colunm = particle.getTransform().getMatrix()[0:3, 2]
        u, v, w = unitCellPlanes

        if (np.dot(colunm, u) > 0) and \
           (np.dot(colunm, v) > 0) and \
           (np.dot(colunm, w) > 0):
            # particle is inside unit cell so nothing needs to be done
            return particle
        else:
            # loop over symmetry matrices
            for i, matrix in enumerate(matrixSet):
                matrix3 = np.array(matrix)
                matrix3 = np.delete(matrix3, -1, 1)
                columPrime = matrix3.dot(colunm)[:3]
                if (np.dot(columPrime, u) > 0) and \
                   (np.dot(columPrime, v) > 0) and \
                   (np.dot(columPrime, w) > 0):
                    matrix = np.dot(matrix, particle.getTransform().getMatrix())
                    t = particle.getTransform()
                    t.setMatrix(matrix)
                    particle.setTransform(t)
                    return particle

            logger.info("Error: something went wrong in moveParticlesInsideUnitCell."
                        " No matrix found to move the particle projection direction inside the unit cell."
                        "       particle id: %s" % particle.getObjId())


def moveParticlesInsideUnitCell(setIN, setOUT, sym=cts.SYM_CYCLIC, n=1):
    """ Move particles projection direction inside the unit cell associated
    to symmetry sym.
    This function (1) gets the symmetry matrices and (2) applies them to
         each projection direction until is inside the unit cell.
    """
    counter = 0  # counter for the number of particles
    totalNumberOfParticles = len(setIN)
    # for each particle in set
    for par in setIN:
        counter += 1
        if counter % 1000 == 0:
            # print something to keep the user engaged
            logger.info(f"{counter}/{totalNumberOfParticles}")

        particle = SymmetryHelper.moveParticleInsideUnitCell(par, sym, n)
        if particle:
            setOUT.append(particle)


def moveParticleInsideUnitCell(particle, matrixSet, unitCellPlanes):
    logger.warning("FOR DEVELOPERS: This method is deprecated. Call SymmetryHelper.moveParticleInsideUnitCell for convenience.")
    SymmetryHelper._moveParticleInsideUnitCell(particle, matrixSet, unitCellPlanes)


def getSymmetryMatrices(sym=cts.SYM_CYCLIC,
                        n=1,
                        circumscribed_radius=1,
                        center=(0, 0, 0),
                        offset=None):
    """ interface between scipion and chimera code
        chimera code uses tuples of tuples as matrices
        but scipion uses np.arrays (lists of lists)
        so let us convert them here.
        Direct use of the classes return 3x3 arrays
        which may be OK in many cases, but not in all
    """
    if sym == cts.SYM_CYCLIC:
        c = Cyclic(n=n,
                   center=center,
                   circumscribed_radius=circumscribed_radius,
                   offset=offset)
        matrices = c.symmetryMatrices()
    elif sym == cts.SYM_DIHEDRAL_X or sym == cts.SYM_DIHEDRAL_Y:
        d = Dihedral(n=n,
                     center=center,
                     circumscribed_radius=circumscribed_radius,
                     offset=offset, sym=sym)
        matrices = d.symmetryMatrices()
    elif sym == cts.SYM_OCTAHEDRAL:
        o = Octahedral(center=center,
                       circumscribed_radius=circumscribed_radius,
                       offset=offset)
        matrices = o.symmetryMatrices()
    elif (sym == cts.SYM_TETRAHEDRAL_222 or
          sym == cts.SYM_TETRAHEDRAL_Z3 or
          sym == cts.SYM_TETRAHEDRAL_Z3R):

        t = Tetrahedral(center=center,
                        circumscribed_radius=circumscribed_radius,
                        sym=sym)
        matrices = t.symmetryMatrices()
    elif (sym == cts.SYM_I222 or sym == cts.SYM_I222r or
          sym == cts.SYM_In25 or sym == cts.SYM_In25r or
          sym == cts.SYM_I2n3 or sym == cts.SYM_I2n3r or
          sym == cts.SYM_I2n5 or sym == cts.SYM_I2n5r):
        matrices = __icosahedralSymmetryMatrices(sym, center)

        # '222'         2-fold symmetry along x, y, and z axes.
        # '222r'        '222' with 90 degree rotation around z.
        # '2n5'         2-fold symmetry along x and 5-fold along z.
        # '2n5r'        '2n5' with 180 degree rotation about y.
        # 'n25'         2-fold symmetry along y and 5-fold along z.
        # 'n25r'        'n25' with 180 degree rotation about x.
        # '2n3'         2-fold symmetry along x and 3-fold along z.
        # '2n3r'        '2n3' with 180 degree rotation about y.

    # convert from 4x3 to 4x4 matrix, Scipion standard
    extraRow = (0., 0., 0., 1.)
    for i in range(len(matrices)):
        matrices[i] += (extraRow,)
    # convert from sets to lists Scipion standard
    return np.array(matrices)


def getUnitCell(sym=cts.SYM_CYCLIC,
                n=1,
                circumscribed_radius=1,
                center=(0, 0, 0),
                offset=None,
                generalize=False):
    """ return vectors normal to the unit cell faces
    """

    if sym == cts.SYM_CYCLIC:
        c = Cyclic(n=n,
                   center=center,
                   circumscribed_radius=circumscribed_radius,
                   offset=offset)
        vectorsEdge, vectorsPlane = c.unitCellPlanes()
        # This function return a three vector planes, but cyclic symmetry only
        # need two, so we duplicate the last one
        vectorsPlane.append(vectorsPlane[-1])
        if DEBUG:
            bildFileName = f'/tmp/C{n}.bild'
            vLabels = ['v1', 'v2', 'v3']
            pLabels = ['p1', 'p2']
    elif sym == cts.SYM_DIHEDRAL_X or sym == cts.SYM_DIHEDRAL_Y:
        logger.info("getUnitCell: %s" % cts.SCIPION_SYM_NAME[sym])
        d = Dihedral(n=n,
                     center=center,
                     circumscribed_radius=circumscribed_radius,
                     offset=offset, sym=sym)
        vectorsEdge, vectorsPlane = d.unitCellPlanes()
        if DEBUG:
            vLabels = ['v1', 'v2', 'eigenvector']
            pLabels = ['p1', 'p2', 'p3']
    elif sym == cts.SYM_OCTAHEDRAL:
        o = Octahedral(center=center,
                       circumscribed_radius=circumscribed_radius)
        vectorsEdge, vectorsPlane = o.unitCellPlanes()
        if DEBUG:
            vLabels = ['v1', 'v2', 'v3']
            pLabels = ['p1', 'p2', 'p3']
    elif (sym == cts.SYM_TETRAHEDRAL_Z3 or
          sym == cts.SYM_TETRAHEDRAL or
          sym == cts.SYM_TETRAHEDRAL_Z3R):
        t = Tetrahedral(center=center,
                        circumscribed_radius=circumscribed_radius,
                        sym=sym)
        vectorsEdge, vectorsPlane = t.unitCellPlanes()
        if DEBUG:
            vLabels = ['v1', 'v2', 'v3']
            pLabels = ['p1', 'p2', 'p3']
    elif (sym == cts.SYM_I222 or sym == cts.SYM_I222r or
          sym == cts.SYM_In25 or sym == cts.SYM_In25r or
          sym == cts.SYM_I2n3 or sym == cts.SYM_I2n3r or
          sym == cts.SYM_I2n5 or sym == cts.SYM_I2n5r):
        i_sym = cts.SCIPION_SYM_NAME[sym][1:]
        i = Icosahedron(center=center,
                        circumscribed_radius=circumscribed_radius,
                        orientation=i_sym)
        vectorsEdge, vectorsPlane = i.unitCellPlanes()
        if DEBUG:
            vLabels = ['v1', 'v2', 'v3']
            pLabels = ['p1', 'p2', 'p3']

    if DEBUG:
        bildFileName = cts.SCIPION_SYM_NAME[sym] + '.bild'
        Chimera.createCoordinateAxisFile(dim=expansionFactor,
                                         bildFileName=bildFileName,
                                         sampling=1)
        colors = ['cyan', 'green', 'magenta', 'navy blue']
        with open(bildFileName, 'a') as f:
            for i, vector in enumerate(vectorsEdge):
                f.write(f'.comment {vLabels[i]}\n')
                f.write(f'.color {colors[i]}\n')
                x = vector[0]*expansionFactor
                y = vector[1]*expansionFactor
                z = vector[2]*expansionFactor
                f.write(f'.arrow 0 0 0  {x} {y} {z} 0.20 0.40 0.75\n')

            for i, vector in enumerate(vectorsPlane):
                f.write(f'.comment {pLabels[i]}\n')
                f.write(f'.color {colors[i]}\n')
                x = vector[0]
                y = vector[1]
                z = vector[2]
                r = expansionFactor/2
                f.write(f'.cone 0 0 0  {x/(expansionFactor*100)} \
                                       {y/(expansionFactor*100)} \
                                       {z/(expansionFactor*100)} {r}\n')
                f.write(f'.arrow 0 0 0  {x*expansionFactor} \
                                        {y*expansionFactor} \
                                        {z*expansionFactor} 0.20 0.40 0.75\n')
        f.close()
    if generalize:
        for i, v in enumerate(vectorsEdge):
            vectorsEdge[i] = np.append(v, 1.)
        for i, v in enumerate(vectorsPlane):
            vectorsPlane[i] = np.append(v, 1.)
    return vectorsEdge, vectorsPlane


def __icosahedralSymmetryMatrices(orientation=cts.SYM_I222, center=(0, 0, 0)):
    if orientation == cts.SYM_I222:
        sym = '222'
    elif orientation == cts.SYM_I222r:
        sym = '222r'
    elif orientation == cts.SYM_In25:
        sym = 'n25'
    elif orientation == cts.SYM_In25r:
        sym = 'n25r'
    elif orientation == cts.SYM_I2n5:
        sym = '2n5'
    elif orientation == cts.SYM_I2n5r:
        sym = '2n5r'
    elif orientation == cts.SYM_I2n3:
        sym = '2n3'
    elif orientation == cts.SYM_I2n3r:
        sym = '2n3r'

    i = Icosahedron(orientation=sym, center=center)
    return list(i.icosahedralSymmetryMatrices())


def __icosahedralUnitCellPlanes(orientation=cts.SYM_I222, center=(0, 0, 0)):
    if orientation == cts.SYM_I222:
        sym = '222'
    elif orientation == cts.SYM_I222r:
        sym = '222r'
    elif orientation == cts.SYM_In25:
        sym = 'n25'
    elif orientation == cts.SYM_In25r:
        sym = 'n25r'
    elif orientation == cts.SYM_I2n5:
        sym = '2n5'
    elif orientation == cts.SYM_I2n5r:
        sym = '2n5r'
    elif orientation == cts.SYM_I2n3:
        sym = '2n3'
    elif orientation == cts.SYM_I2n3r:
        sym = '2n3r'

    i = Icosahedron(orientation=sym, center=center)
    return list(i.unitCellPlanes())


class Cyclic(object):
    """cyclic class.
        Allow to compute symmetry matrices, vertices,
        symmetry axis and unit cell planes
    """

    def __init__(self, n, circumscribed_radius=1, center=(
            0, 0, 0), offset=None):
        """
        :Parameters:
            n: int order of symmetry
            circumscribed_radius: float radius of the circumscribed sphere
            center: tuple of 3 floats, center of the circumscribed sphere
            offset: float, angle in degrees to rotate the symmetry axis
                    it modifies the unit cell but not the symmetry matrices
            """
        self.n = n
        self.circumscribed_radius = circumscribed_radius
        self.center = center
        self.matrices = None
        self.offset = offset

    def symmetryMatrices(self):
        """ get Matrices for cyclic symmetry of order n
        """
        if self.matrices is not None:
            return self.matrices
        self.matrices = []
        for k in range(self.n):
            a = 2*pi * np.float32(k) / self.n
            c = cos(a)
            s = sin(a)
            tf = ((c, -s, 0, 0),
                  (s, c, 0, 0),
                  (0, 0, 1, 0))
            self.matrices.append(tf)
        self.matrices = _recenterSymmetries(self.matrices, self.center)
        return self.matrices

    def unitCellPlanes(self):
        """ get planes that define a unit cell for cyclic symmetry of order n
        """
        if self.offset is None:
            self.offset = pi/(self.n)
        _ = self.symmetryMatrices()
        # these three vectors are enges of the unit cell
        angle = -2*pi/self.n
        a1 = self.offset
        a2 = angle + self.offset
        c1 = cos(a1)
        s1 = sin(a1)
        c2 = cos(a2)
        s2 = sin(a2)
        v1 = np.array([c1, s1, 0.])
        v2 = np.array([c2, s2, 0.])
        v3 = np.array([0., 0., 1.])  # keep this dim 3 instead of dim 4

        # cross product of v1/v2 with eigenvetor
        # these two vectors are normal to the planes that define the unit cell
        plane1 = np.cross(v1, v3)
        plane2 = np.cross(v3, v2)
        if DEBUG:
            logger.info("v1 %s" % v1)
            logger.info("v2 %s" % v2)
            logger.info("v3 %s" % v3)
            logger.info("plane1 %s" % plane1)
            logger.info("plane2 %s" % plane2)
        c = self.circumscribed_radius
        return [v1*c, v2*c, v3*c], [plane1*c, plane2*c]


class Dihedral(Cyclic):
    """dihedral class.
        Allow to compute symmetry matrices,  vertices,
        symmetry axis and unit cell planes
    """
    def __init__(self, n, circumscribed_radius=1, center=(
            0, 0, 0), offset=None, sym=cts.SYM_DIHEDRAL_X):
        """
        :Parameters:
            n: int order of symmetry
            circumscribed_radius: float radius of the circumscribed sphere
            center: tuple of 3 floats, center of the circumscribed sphere
            offset: float, angle in degrees to rotate the symmetry axis
                    it modifies the unit cell but not the symmetry matrices
            """
        super().__init__(n=n,
                         circumscribed_radius=circumscribed_radius,
                         center=center,
                         offset=offset)
        self.sym = sym

    def symmetryMatrices(self):
        clist = super().symmetryMatrices()
        if self.sym == cts.SYM_DIHEDRAL_X:
            # matrix that reflect in the x axis
            reflect = ((1, 0, 0, 0), (0, -1, 0, 0), (0, 0, -1, 0))
        else:
            logger.info("Be carefull untested code")
            # matrix that reflect in the y axis
            reflect = ((-1, 0, 0, 0), (0, 1, 0, 0), (0, 0, -1, 0))

        tflist = _matrixProducts([_identityMatrix(), reflect], clist)
        tflist = _recenterSymmetries(tflist, self.center)
        return tflist

    def unitCellPlanes(self):
        """ get planes that define a unit cell for dihedral symmetry of order n
        """
        if self.offset is None:
            self.offset = pi/(self.n) + pi/2
        vectorsEdge, vectorsPlane = super().unitCellPlanes()
        vectorsPlane.append(np.array([0., 0., self.circumscribed_radius]))
        if DEBUG:
            logger.info("v1 %s" % vectorsEdge[0])
            logger.info("v2 %s" % vectorsEdge[1])
            logger.info("eigenvector %s" % vectorsEdge[2])
            logger.info("plane1 %s" % vectorsPlane[0])
            logger.info("plane2 %s" % vectorsPlane[1])
            logger.info("plane3 %s" % vectorsPlane[2])

        return vectorsEdge, vectorsPlane

    def coordinateSystemTransform(self, origSym, targetSym):
        logger.info("========================%s->%s" % ( cts.SCIPION_SYM_NAME[origSym], cts.SCIPION_SYM_NAME[targetSym]))
        if origSym == targetSym:
            return np.identity(4)
        elif origSym == cts.SYM_DIHEDRAL_X and targetSym == cts.SYM_DIHEDRAL_Y:
            return ((0., -1., 0., 0.), (1, 0., 0., 0.), (0., 0., 1., 0.))
        elif origSym == cts.SYM_DIHEDRAL_Y and targetSym == cts.SYM_DIHEDRAL_X:
            return ((0., 1., 0., 0.), (-1, 0., 0., 0.), (0., 0., 1., 0.))
        else:
            raise Exception("unknown dihedral symmetry pair: %s %s" % (
                cts.SCIPION_SYM_NAME[origSym],
                cts.SCIPION_SYM_NAME[targetSym]))


class Tetrahedral(object):
    """Tetrahedral class.
        Allow to compute symmetry matrices,  vertices,
        symmetry axis and unit cell planes
    """

    def __init__(self, circumscribed_radius=1, center=(
            0, 0, 0), sym=cts.SYM_TETRAHEDRAL_Z3):
        """
        :Parameters:
            n: int order of symmetry
            circumscribed_radius: float radius of the circumscribed sphere
            center: tuple of 3 floats, center of the circumscribed sphere
            """
        self.circumscribed_radius = circumscribed_radius
        self.center = center
        self.matrices = None
        self.sym = sym

    def symmetryMatrices(self):
        """
        identity
        4 * rotation by 120 clockwise (seen from a vertex):
                            (234), (143), (412), (321)
        4 * rotation by 120 counterclockwise (ditto)
        3 * rotation by 180
         """
        # simmetries are rotations around theses axis by these angles
        self.aa = (((0, 0, 1), 0), ((1, 0, 0), 180),  # 0 1
                   ((0, 1, 0), 180), ((0, 0, 1), 180),  # 2 3
                   ((1, 1, 1), 120), ((1, 1, 1), 240),  # 4 5
                   ((-1, -1, 1), 120), ((-1, -1, 1), 240),  # 6 7
                   ((-1, 1, -1), 120), ((-1, 1, -1), 240),  # 8 9
                   ((1, -1, -1), 120),  ((1, -1, -1), 240))  # 10 11

        syms = [_rotationTransform(axis, angle) for axis, angle in self.aa]

        if self.sym == cts.SYM_TETRAHEDRAL_Z3R:
            # convention, 3-fold on z, 3-fold in yz plane sign(y) = sign(z)
            self.tf = _multiplyMatrices(
                     _rotationTransform((0, 0, 1), -45.0),
                     _rotationTransform((1, 0, 0),
                                        -acos(1 / sqrt(3)) * 180 / pi))
            syms = _coordinateTransformList(syms, self.tf)
        elif self.sym == cts.SYM_TETRAHEDRAL_Z3:
            # convention, 3-fold on z, 3-fold in yz plane sign(y) != sign(z)
            tfaux = _multiplyMatrices(
                     _rotationTransform((0, 0, 1), -45.0),
                     _rotationTransform((1, 0, 0),
                                        -acos(1 / sqrt(3)) * 180 / pi)
            )
            self.tf = _multiplyMatrices(
                     tfaux,
                     _reflectMatrix((0, 0, 1))
            )
            syms = _coordinateTransformList(syms, self.tf)

        syms = _recenterSymmetries(syms, self.center)
        return syms

    def unitCellPlanes(self):
        """ get planes that define a unit cell for tetrahedral symmetry of
            order n
        """
        logger.info("Unit cell sym %s" % cts.SCIPION_SYM_NAME[self.sym])
        _ = self.symmetryMatrices()
        # four corners tetahedron
        v0 = _normalizeVector(np.array(self.aa[4][0])) *\
            self.circumscribed_radius
        v1 = _normalizeVector(np.array(self.aa[6][0])) *\
            self.circumscribed_radius
        v2 = _normalizeVector(np.array(self.aa[8][0])) *\
            self.circumscribed_radius
        v3 = _normalizeVector(np.array(self.aa[10][0])) *\
            self.circumscribed_radius

        if self.sym == cts.SYM_TETRAHEDRAL_Z3 or\
           self.sym == cts.SYM_TETRAHEDRAL_Z3R:
            self.tf = _invertMatrix(self.tf)
            if self.sym == cts.SYM_TETRAHEDRAL_Z3:
                _3fold_1 = np.dot(self.tf, np.append(v0, 1))
                _3fold_2 = np.dot(self.tf, np.append(v1, 1))
                _3fold_3 = np.dot(self.tf, np.append(v3, 1))
            else:
                _3fold_1 = np.dot(self.tf, np.append(v1, 1))
                _3fold_2 = np.dot(self.tf, np.append(v0, 1))
                _3fold_3 = np.dot(self.tf, np.append(v3, 1))
        else:
            _3fold_1 = v1
            _3fold_2 = v0
            _3fold_3 = v2

        # colors = ['cyan', 'green', 'magenta', 'navy blue']
        vp = np.add(_3fold_1, _3fold_2)
        vp = _normalizeVector(np.add(vp, _3fold_3))

        plane1 = _normalizeVector(np.cross(_3fold_1, _3fold_2))  # cyan
        plane2 = _normalizeVector(np.cross(_3fold_2, vp))  # green
        plane3 = _normalizeVector(np.cross(vp, _3fold_1))  # magenta
        if DEBUG:
            logger.info("v0 %s" % _3fold_1)
            logger.info("v1 %s" % _3fold_2)
            logger.info("v2 %s" % vp)

            logger.info("plane1 %s" % plane1)
            logger.info("plane2 %s" % plane2)
            logger.info("plane3 %s" % plane3)
        return [_3fold_1, vp, _3fold_2], [plane1, plane2, plane3]

    def coordinateSystemTransform(self, origSym, targetSym):
        if origSym == targetSym:
            return np.identity(4)

        elif (origSym == cts.SYM_TETRAHEDRAL_222 and
              targetSym == cts.SYM_TETRAHEDRAL_Z3 or
              origSym == cts.SYM_TETRAHEDRAL_222 and
              targetSym == cts.SYM_TETRAHEDRAL_Z3R):
            logger.info("========================222 -> ")
            matrix = _multiplyMatrices(
                     _rotationTransform((1, 0, 0),
                                        -acos(1 / sqrt(3)) * 180 / pi),
                     _rotationTransform((0, 0, 1),  45.0)
            )
            # matrix = _multiplyMatrices(
            #          matrix,
            #          _reflectMatrix((0, 0, 1))
            # )
            # matrix = _invertMatrix(matrix)
            if targetSym == cts.SYM_TETRAHEDRAL_Z3:
                return matrix
            else:
                matrix2 = _reflectMatrix((1, 0, 0))
                return _multiplyMatrices(matrix2, matrix)

        elif (origSym == cts.SYM_TETRAHEDRAL_Z3R and
              targetSym == cts.SYM_TETRAHEDRAL_Z3) or\
             (origSym == cts.SYM_TETRAHEDRAL_Z3 and
              targetSym == cts.SYM_TETRAHEDRAL_Z3R):
            matrix = _reflectMatrix((1, 0, 0))
            return matrix

        elif (origSym == cts.SYM_TETRAHEDRAL_Z3 and
              targetSym == cts.SYM_TETRAHEDRAL_222) or\
             (origSym == cts.SYM_TETRAHEDRAL_Z3R and
              targetSym == cts.SYM_TETRAHEDRAL_222):
            matrix = self.coordinateSystemTransform(targetSym, origSym)
            return _invertMatrix(matrix)

        else:
            raise Exception("unknown Tetrahedral symmetry pair: %s %s" % (
                cts.SCIPION_SYM_NAME[origSym],
                cts.SCIPION_SYM_NAME[targetSym]))


class Octahedral(object):
    """Octahedral class.
        Allow to compute symmetry matrices,
        symmetry axis and unit cell planes of an octahedral symmetry.
        4-folds along x, y, z axes.
    """
    def __init__(self, circumscribed_radius=1, center=(
            0, 0, 0), offset=None):
        """
        :Parameters:
            circumscribed_radius: float radius of the circumscribed sphere
            center: tuple of 3 floats, center of the circumscribed sphere
            offset: float, angle in degrees to rotate the symmetry axis
                    it modifies the unit cell but not the symmetry matrices
            """
        self.circumscribed_radius = circumscribed_radius
        self.center = center
        self.matrices = None
        self.offset = offset

    def symmetryMatrices(self):
        """ get Matrices for cyclic symmetry of order n
        """
        self.c4 = (((0, 0, 1), 0),
                   ((0, 0, 1), 90),
                   ((0, 0, 1), 180),
                   ((0, 0, 1), 270))
        self.cube = (((1, 0, 0), 0),
                     ((1, 0, 0), 90),
                     ((1, 0, 0), 180),
                     ((1, 0, 0), 270),
                     ((0, 1, 0), 90),
                     ((0, 1, 0), 270))
        c4syms = [_rotationTransform(axis, angle)
                  for axis, angle in self.c4]
        cubesyms = [_rotationTransform(axis, angle)
                    for axis, angle in self.cube]
        syms = _matrixProducts(cubesyms, c4syms)
        self.matrices = _recenterSymmetries(syms, self.center)
        return self.matrices

    def unitCellPlanes(self):
        """ get planes that define a unit cell for an octahedral symmetry
        """
        _ = self.symmetryMatrices()
        # four corners tetahedron
        _4fold_1 = _normalizeVector(
            np.array([0., 0., 1.])) * self.circumscribed_radius
        _3fold_1 = _normalizeVector(
            np.array([1., 1., 1.])) * self.circumscribed_radius
        _3fold_2 = _normalizeVector(
            np.array([-1., 1., 1.])) * self.circumscribed_radius

        plane1 = _normalizeVector(
            np.cross(_3fold_2, _4fold_1))  # cyan
        plane2 = _normalizeVector(
            np.cross(_3fold_1, _3fold_2))  # green
        plane3 = _normalizeVector(
            np.cross(_4fold_1, _3fold_1))  # magenta
        if DEBUG:
            logger.info("v0 %s" % _4fold_1)
            logger.info("v1 %s" % _3fold_2)
            logger.info("v2 %s" % _3fold_1)

            logger.info("plane1 %s" % plane1)
            logger.info("plane2 %s" % plane2)
            logger.info("plane3 %s" % plane3)
        return [_4fold_1, _3fold_2, _3fold_1], [plane1, plane2, plane3]


# TODO: Why is this a global variable instead of
# a class variable?
icos_matrices = {}  # Maps orientation name to 60 matrices.


class Icosahedron(object):
    """Icosahedron class.
        Allow to compute symmetry matrices, icosahedron vertices,
         symmetry axis and unit cell planes
    """
    def __init__(self, circumscribed_radius=1, orientation='222', center=(
            0, 0, 0)):
        """point = np.array([0, 1, PHI]"""
        self.circumscribed_radius = circumscribed_radius
        self.orientation = orientation
        self.center = center
        # Triangle edge length of unit icosahedron.
        self.e = sqrt(2 - 2 / sqrt(5))
        self.vertices = None  # icosahedron vertices
        self.triangles = None  # icosahedron faces
        self._3foldAxis = []
        self._2foldAxis = []
        self.edges = None  # icosahedron edges
        # ---------------------------------------------------------------------------------------------
        # Coordinates systems.
        # '222'         2-fold symmetry along x, y, and z axes.
        # '222r'        '222' with 90 degree rotation around z.
        # '2n5'         2-fold symmetry along x and 5-fold along z.
        # '2n5r'        '2n5' with 180 degree rotation about y.
        # 'n25'         2-fold symmetry along y and 5-fold along z.
        # 'n25r'        'n25' with 180 degree rotation about x.
        # '2n3'         2-fold symmetry along x and 3-fold along z.
        # '2n3r'        '2n3' with 180 degree rotation about y.
        #
        self.coordinate_system_names = ('222', '222r', '2n5',
                                        '2n5r', 'n25', 'n25r', '2n3', '2n3r')
        self.icosahedronEdgeLength = \
            self.icosahedronEdgeLength(circumscribed_radius)
        self.angle23, self.angle25, self.angle35 = self.icosahedronAngles()
        self.icosahedronGeometry()

    # ---------------------------------------------------------------------------------------------
    # 60 icosahedral symmetry matrices.
    #
    def icosahedralSymmetryMatrices(self):
        t = self.icosahedralMatrixTable()
        tflist = _recenterSymmetries(t.get(self.orientation, None),
                                     self.center)
        return tflist

    # -------------------------------------------------------------------------
    # Edge length of icosahedron with a certain radio of the circumscribed
    # sphere:
    # According to Radio of the circumscribed sphere = 1 (vertices),
    #  the icosahedronEdgeLength should be 1.0515

    def icosahedronEdgeLength(self, circumscribed_radius):
        return 4 * circumscribed_radius / sqrt(10 + 2 * sqrt(5))

    # -------------------------------------------------------------------------
    # Matrices for mapping between different icosahedron coordinate frames.
    #
    def coordinateSystemTransform(self, from_cs, to_cs):

        self.cst = {}
        if self.cst:
            return self.cst[(from_cs, to_cs)]

        transform = self.cst

        s25 = self.e / 2  # Sin/Cos for angle between 2-fold and 5-fold axis
        c25 = sqrt(1 - s25 * s25)
        # Sin/Cos for angle between 3-fold and 5-fold axis
        s35 = self.e / sqrt(3)
        c35 = sqrt(1 - s35 * s35)

        transform[('2n5', '222')] = ((1, 0, 0, 0),
                                     (0, c25, -s25, 0),
                                     (0, s25, c25, 0))
        transform[('2n5', '2n3')] = ((1, 0, 0, 0),
                                     (0, c35, s35, 0),
                                     (0, -s35, c35, 0))

        # Axes permutations.
        # 90 degree rotation about z
        transform[('222', '222r')] = ((0, 1, 0, 0),
                                      (-1, 0, 0, 0),
                                      (0, 0, 1, 0))
        # 180 degree rotation about y
        transform[('2n3', '2n3r')] = \
            transform[('2n5', '2n5r')] = ((-1, 0, 0, 0),
                                          (0, 1, 0, 0),
                                          (0, 0, -1, 0))
        # 180 degree rotation about x
        transform[('n25', 'n25r')] = ((1, 0, 0, 0),
                                      (0, -1, 0, 0),
                                      (0, 0, -1, 0))
        # x <-> y and z -> -z
        transform[('n25', '2n5')] = ((0, 1, 0, 0),
                                     (1, 0, 0, 0),
                                     (0, 0, -1, 0))

        # Extend to all pairs of transforms.
        tlist = []
        while len(transform) > len(tlist):

            tlist = list(transform.keys())

            # Add inverse transforms
            for key in tlist:
                f = key[0]
                t = key[1]
                if not (t, f) in transform:
                    transform[(t, f)] = _transposeMatrix(transform[(f, t)])

            # Use transitivity
            for key1 in tlist:
                f1 = key1[0]
                t1 = key1[1]
                for key2 in tlist:
                    f2 = key2[0]
                    t2 = key2[1]
                    if f2 == t1 and f1 != t2 and not (f1, t2) in transform:
                        transform[(f1, t2)] = _multiplyMatrices(transform[(
                            f2, t2)], transform[(f1, t1)])

        i = _identityMatrix()
        for s in self.coordinate_system_names:
            transform[(s, s)] = i

        return transform[(from_cs, to_cs)]

    # ---------------------------------------------------------------------------------------
    # Compute icosahedral transformation matrices for different
    # coordinate systems.
    #

    def icosahedralMatrixTable(self):
        global icos_matrices
        if icos_matrices:
            return icos_matrices

        c = cos(2 * pi / 5)  # .309016994
        c2 = cos(4 * pi / 5)  # -.809016994

        icos_matrices['222'] = (

            ((1.0, 0.0, 0.0, 0.0),
             (0.0, 1.0, 0.0, 0.0),
             (0.0, 0.0, 1.0, 0.0)),

            ((c2, -0.5, c, 0.0),
             (-0.5, c, c2, 0.0),
             (c, c2, -0.5, 0.0)),

            ((0.0, 1.0, 0.0, 0.0),
             (0.0, 0.0, -1.0, 0.0),
             (-1.0, 0.0, 0.0, 0.0)),

            ((-c2, -0.5, -c, 0.0),
             (-0.5, -c, c2, 0.0),
             (c, -c2, -0.5, 0.0)),

            ((0.5, c, c2, 0.0),
             (-c, c2, -0.5, 0.0),
             (c2, 0.5, -c, 0.0)),

            ((-c, c2, -0.5, 0.0),
             (c2, 0.5, -c, 0.0),
             (0.5, c, c2, 0.0)),

            ((c2, 0.5, -c, 0.0),
             (0.5, c, c2, 0.0),
             (-c, c2, -0.5, 0.0)),

            ((c2, -0.5, -c, 0.0),
             (0.5, -c, c2, 0.0),
             (c, c2, 0.5, 0.0)),

            ((-c, -c2, -0.5, 0.0),
             (c2, -0.5, -c, 0.0),
             (-0.5, c, -c2, 0.0)),

            ((0.5, -c, c2, 0.0),
             (-c, -c2, -0.5, 0.0),
             (-c2, 0.5, c, 0.0)),

            ((0.0, 0.0, -1.0, 0.0),
             (-1.0, 0.0, 0.0, 0.0),
             (0.0, 1.0, 0.0, 0.0)),

            ((-0.5, -c, c2, 0.0),
             (c, -c2, -0.5, 0.0),
             (-c2, -0.5, -c, 0.0)),

            ((-0.5, c, c2, 0.0),
             (c, c2, -0.5, 0.0),
             (c2, -0.5, c, 0.0)),

            ((-c, c2, -0.5, 0.0),
             (-c2, -0.5, c, 0.0),
             (-0.5, -c, -c2, 0.0)),

            ((c2, 0.5, -c, 0.0),
             (-0.5, -c, -c2, 0.0),
             (c, -c2, 0.5, 0.0)),

            ((0.5, c, c2, 0.0),
             (c, -c2, 0.5, 0.0),
             (-c2, -0.5, c, 0.0)),

            ((-0.5, c, c2, 0.0),
             (-c, -c2, 0.5, 0.0),
             (-c2, 0.5, -c, 0.0)),

            ((0.0, 0.0, -1.0, 0.0),
             (1.0, 0.0, 0.0, 0.0),
             (0.0, -1.0, 0.0, 0.0)),

            ((-0.5, -c, c2, 0.0),
             (-c, c2, 0.5, 0.0),
             (c2, 0.5, c, 0.0)),

            ((0.0, -1.0, 0.0, 0.0),
             (0.0, 0.0, 1.0, 0.0),
             (-1.0, 0.0, 0.0, 0.0)),

            ((c2, 0.5, c, 0.0),
             (0.5, c, -c2, 0.0),
             (c, -c2, -0.5, 0.0)),

            ((-c2, 0.5, -c, 0.0),
             (0.5, -c, -c2, 0.0),
             (c, c2, -0.5, 0.0)),

            ((-c, -c2, -0.5, 0.0),
             (-c2, 0.5, c, 0.0),
             (0.5, -c, c2, 0.0)),

            ((0.5, -c, c2, 0.0),
             (c, c2, 0.5, 0.0),
             (c2, -0.5, -c, 0.0)),

            ((c2, -0.5, -c, 0.0),
             (-0.5, c, -c2, 0.0),
             (-c, -c2, -0.5, 0.0)),

            ((-c, c2, 0.5, 0.0),
             (c2, 0.5, c, 0.0),
             (-0.5, -c, c2, 0.0)),

            ((-c, -c2, 0.5, 0.0),
             (-c2, 0.5, -c, 0.0),
             (-0.5, c, c2, 0.0)),

            ((1.0, 0.0, 0.0, 0.0),
             (0.0, -1.0, 0.0, 0.0),
             (0.0, 0.0, -1.0, 0.0)),

            ((c, -c2, -0.5, 0.0),
             (-c2, -0.5, -c, 0.0),
             (-0.5, -c, c2, 0.0)),

            ((c, c2, -0.5, 0.0),
             (c2, -0.5, c, 0.0),
             (-0.5, c, c2, 0.0)),

            ((-1.0, 0.0, 0.0, 0.0),
             (0.0, 1.0, 0.0, 0.0),
             (0.0, 0.0, -1.0, 0.0)),

            ((-c2, 0.5, -c, 0.0),
             (-0.5, c, c2, 0.0),
             (-c, -c2, 0.5, 0.0)),

            ((0.0, -1.0, 0.0, 0.0),
             (0.0, 0.0, -1.0, 0.0),
             (1.0, 0.0, 0.0, 0.0)),

            ((c2, 0.5, c, 0.0),
             (-0.5, -c, c2, 0.0),
             (-c, c2, 0.5, 0.0)),

            ((-0.5, -c, -c2, 0.0),
             (-c, c2, -0.5, 0.0),
             (-c2, -0.5, c, 0.0)),

            ((c, -c2, 0.5, 0.0),
             (c2, 0.5, -c, 0.0),
             (-0.5, -c, -c2, 0.0)),

            ((-c2, -0.5, c, 0.0),
             (0.5, c, c2, 0.0),
             (c, -c2, 0.5, 0.0)),

            ((-c2, 0.5, c, 0.0),
             (0.5, -c, c2, 0.0),
             (-c, -c2, -0.5, 0.0)),

            ((c, c2, 0.5, 0.0),
             (c2, -0.5, -c, 0.0),
             (0.5, -c, c2, 0.0)),

            ((-0.5, c, -c2, 0.0),
             (-c, -c2, -0.5, 0.0),
             (c2, -0.5, -c, 0.0)),

            ((0.0, 0.0, 1.0, 0.0),
             (-1.0, 0.0, 0.0, 0.0),
             (0.0, -1.0, 0.0, 0.0)),

            ((0.5, c, -c2, 0.0),
             (c, -c2, -0.5, 0.0),
             (c2, 0.5, c, 0.0)),

            ((0.5, -c, -c2, 0.0),
             (c, c2, -0.5, 0.0),
             (-c2, 0.5, -c, 0.0)),

            ((c, -c2, 0.5, 0.0),
             (-c2, -0.5, c, 0.0),
             (0.5, c, c2, 0.0)),

            ((-c2, -0.5, c, 0.0),
             (-0.5, -c, -c2, 0.0),
             (-c, c2, -0.5, 0.0)),

            ((-0.5, -c, -c2, 0.0),
             (c, -c2, 0.5, 0.0),
             (c2, 0.5, -c, 0.0)),

            ((0.5, -c, -c2, 0.0),
             (-c, -c2, 0.5, 0.0),
             (c2, -0.5, c, 0.0)),

            ((0.0, 0.0, 1.0, 0.0),
             (1.0, 0.0, 0.0, 0.0),
             (0.0, 1.0, 0.0, 0.0)),

            ((0.5, c, -c2, 0.0),
             (-c, c2, 0.5, 0.0),
             (-c2, -0.5, -c, 0.0)),

            ((0.0, 1.0, 0.0, 0.0),
             (0.0, 0.0, 1.0, 0.0),
             (1.0, 0.0, 0.0, 0.0)),

            ((-c2, -0.5, -c, 0.0),
             (0.5, c, -c2, 0.0),
             (-c, c2, 0.5, 0.0)),

            ((c2, -0.5, c, 0.0),
             (0.5, -c, -c2, 0.0),
             (-c, -c2, 0.5, 0.0)),

            ((c, c2, 0.5, 0.0),
             (-c2, 0.5, c, 0.0),
             (-0.5, c, -c2, 0.0)),

            ((-0.5, c, -c2, 0.0),
             (c, c2, 0.5, 0.0),
             (-c2, 0.5, c, 0.0)),

            ((-c2, 0.5, c, 0.0),
             (-0.5, c, -c2, 0.0),
             (c, c2, 0.5, 0.0)),

            ((c, -c2, -0.5, 0.0),
             (c2, 0.5, c, 0.0),
             (0.5, c, -c2, 0.0)),

            ((c, c2, -0.5, 0.0),
             (-c2, 0.5, -c, 0.0),
             (0.5, -c, -c2, 0.0)),

            ((-1.0, 0.0, 0.0, 0.0),
             (0.0, -1.0, 0.0, 0.0),
             (0.0, 0.0, 1.0, 0.0)),

            ((-c, c2, 0.5, 0.0),
             (-c2, -0.5, -c, 0.0),
             (0.5, c, -c2, 0.0)),

            ((-c, -c2, 0.5, 0.0),
             (c2, -0.5, c, 0.0),
             (0.5, -c, -c2, 0.0)),

        )

        for cs in self.coordinate_system_names:
            if cs != '222':
                t = self.coordinateSystemTransform(cs, '222')
                tinv = self.coordinateSystemTransform('222', cs)
                icos_matrices[cs] = [_multiplyMatrices(tinv, m, t)
                                     for m in icos_matrices['222']]
        return icos_matrices

    # -----------------------------------------------------------------------------

    def icosahedronAngles(self):
        # Angles between 2-fold, 3-fold and 5-fold
        # symmetry axes of an icosahedron.
        angle25 = asin(self.e/2)
        angle35 = asin(self.e/sqrt(3))
        angle23 = asin(self.e/(2*sqrt(3)))
        return angle23, angle25, angle35

    def icosahedronGeometry(self):

        a = 2 * self.angle25  # Angle spanned by edge from center
        # 5-fold symmetry axis along z
        c5 = cos(2 * pi / 5)
        s5 = sin(2 * pi / 5)
        tf5 = ((c5, -s5, 0, 0),
               (s5, c5, 0, 0),
               (0, 0, 1, 0))

        # 2-fold symmetry axis along x
        tf2 = ((1, 0, 0, 0),
               (0, -1, 0, 0),
               (0, 0, -1, 0))

        p = tuple(map(operator.add, self.center,
                      (0, 0, self.circumscribed_radius)))
        p50 = tuple(map(operator.add, self.center,
                        (0, self.circumscribed_radius * sin(a),
                         self.circumscribed_radius * cos(a))))
        p51 = _applyMatrix(tf5, p50)
        p52 = _applyMatrix(tf5, p51)
        p53 = _applyMatrix(tf5, p52)
        p54 = _applyMatrix(tf5, p53)
        self.vertices = [p, p50, p51, p52, p53, p54]
        self.vertices.extend([_applyMatrix(tf2, q) for q in self.vertices])
        if self.orientation != '2n5':
            self.tf = self.coordinateSystemTransform('2n5', self.orientation)
            self.vertices = [_applyMatrix(self.tf, p) for p in self.vertices]

            #
            # Vertex numbering
            #
            #  Top   1          Bottom
            #      2   5        9   10
            #        0            6
            #      3   4        8   11
            #                     7
        # 20 triangles composing icosahedron.
        #
        self.triangles = ((0, 1, 2), (0, 2, 3),
                          (0, 3, 4), (0, 4, 5),
                          (0, 5, 1), (6, 7, 8),
                          (6, 8, 9), (6, 9, 10),
                          (6, 10, 11), (6, 11, 7),
                          (1, 9, 2), (2, 9, 8),
                          (2, 8, 3), (3, 8, 7),
                          (3, 7, 4), (4, 7, 11),
                          (4, 11, 5), (5, 11, 10),
                          (5, 10, 1), (1, 10, 9))

        for triangle in self.triangles:
            self._3foldAxis.append([(item1 + item2 + item3) / 3.
                                    for item1,  item2,  item3 in zip(
                                        self.vertices[triangle[0]],
                                        self.vertices[triangle[1]],
                                        self.vertices[triangle[2]])])

        self.edges = ((0, 1), (0, 2), (0, 3), (0, 4), (0, 5),
                      (1, 2), (2, 3), (3, 4), (4, 5), (5, 1),
                      (6, 7), (6, 8), (6, 9), (6, 10), (6, 11),
                      (1, 9), (2, 8), (2, 9), (3, 7), (3, 8),
                      (7, 8), (8, 9), (9, 10), (10, 11), (11, 7),
                      (4, 7), (4, 11), (5, 10), (5, 11), (1, 10))

        for edge in self.edges:
            self._2foldAxis.append([(item1 + item2) / 2.
                                    for item1, item2 in zip(
                                        self.vertices[edge[0]],
                                        self.vertices[edge[1]])
                                    ])

    def getVertices(self):
        return self.vertices

    def getTriangles(self):
        return self.triangles

    def getEdges(self):
        return self.edges

    def get3foldAxis(self):
        return self._3foldAxis

    def get2foldAxis(self):
        return self._2foldAxis

    def unitCellPlanes(self):
        """ get planes that define a unit cell for an octahedral symmetry
        """
        if self.orientation == '222':
            # Unit cell in triangle 11 -> (2,8,9)
            _3f = 11; _5f1 = 2; _5f2 = 8  # noqa
        elif self.orientation == '222r':
            # Unit cell in triangle 0 -> (0,1,2)
            _3f = 0; _5f1 = 1; _5f2 = 0  # noqa
        elif self.orientation == 'n25':
            # Unit cell in triangle 9 -> (6,11,7)
            _3f = 9; _5f1 = 6; _5f2 = 7  # noqa
        elif self.orientation == 'n25r':
            # Unit cell in triangle 14 -> (3,7,4)
            _3f = 14; _5f1 = 3; _5f2 = 4  # noqa
        elif self.orientation == '2n3':  # TODO: Untested
            _3f = 11; _5f1 = 8; _5f2 = 2  # noqa
        elif self.orientation == '2n3r':  # TODO: Untested
            _3f = 11; _5f1 = 2; _5f2 = 8  # noqa
        elif self.orientation == '2n5':  # TODO: Untested
            _3f = 11; _5f1 = 8; _5f2 = 2  # noqa
        elif self.orientation == '2n5r':  # TODO: Untested
            _3f = 11; _5f1 = 2; _5f2 = 8  # noqa
        _3fold = self._3foldAxis[_3f]
        _5fold_1 = self.getVertices()[_5f1]
        _5fold_2 = self.getVertices()[_5f2]

        plane1 = _normalizeVector(np.cross(_5fold_1, _3fold))  # cyan
        plane2 = _normalizeVector(np.cross(_3fold, _5fold_2))  # magenta
        plane3 = _normalizeVector(np.cross(_5fold_2, _5fold_1))  # green
        if DEBUG:
            logger.info("v0 %s" % _3fold)
            logger.info("v1 %s" % _5fold_1)
            logger.info("v2 %s" % _5fold_2)

            logger.info("plane1 %s" % plane1)
            logger.info("plane2 %s" % plane2)
            logger.info("plane3 %s" % plane3)
        return [_3fold, _5fold_1, _5fold_2], [plane1, plane2, plane3]



# **************************************************************************
# *
# * Authors:     Airen Zaldivar Peraza (azaldivar@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
Protocol base classes related to 2D processing.
"""
import pyworkflow.protocol.params as params
import pyworkflow.utils as pwutils

import pwem.objects as emobj
from .protocol import EMProtocol


class Prot2D(EMProtocol):
    pass


class ProtAlign2D(Prot2D):
    """ This class will serve as a base for all protocols that align a set of 
        2D images.
        All Align protocols receive as input:
        A set of particles
        and will allow the option to generate the aligned particles.
    """
    def _defineParams(self, form):
        form.addSection(label='Input')
        
        form.addParam('inputParticles', params.PointerParam, important=True,
                      label=pwutils.Message.LABEL_INPUT_PART, pointerClass='SetOfParticles')
        # Hook that should be implemented in subclasses
        self._defineAlignParams(form)
        
    def _defineAlignParams(self, form):
        """ This method should be implemented by subclasses
        to add other parameter relatives to the specific align protocol."""
        pass 
    

class ProtCreateMask2D(Prot2D):
    """ For those protocols who create mask as output. """
    pass


class ProtClassify2D(Prot2D):
    pass


class ProtAnalysis2D(Prot2D):
    pass


# ------------ Some testing protocols -----------------------------------------

class ProtEvenClassify2D:  # ProtClassify2D):
    """ This is a simple classification protocol for testing purposes.
    It will iterate over the images and assign each one to a different
    class. 
    """
    def _defineParams(self, form):
        form.addSection(label='Input')
        
        form.addParam('inputParticles', params.PointerParam, important=True,
                      label=pwutils.Message.LABEL_INPUT_PART, pointerClass='SetOfParticles')
        form.addParam('numberOfClasses', params.IntParam, default=8,
                      label="Number of classes", 
                      help="Select into how many classes do you want to classify your images.")

    def _insertAllSteps(self):
        self._insertFunctionStep('createOutputStep')
        
    def createOutputStep(self):
        """ This single step will do the fullish classification
        and create the output classes.
        """
        classes = self._createSetOfClasses2D()
        n = self.numberOfClasses.get()
        
        for classNumber in range(1, n+1):
            classes.append(emobj.Class2D())
        
        for img in self.inputParticles.get():
            classNumber = (img.getObjId() % n) + 1
            classes[classNumber].append(img)
            
        for cls in classes:
            classes.update(cls)
            
        classes.write()
            
        self._defineOutputs(outputClasses=classes)
